# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from .v1.client import ManageClient as ManageClientLatest
from .v1.async_client import AsyncManageClient as AsyncManageClientLatest

# input
from .v1.options import (
    ProjectOptions as ProjectOptionsLatest,
    KeyOptions as KeyOptionsLatest,
    ScopeOptions as ScopeOptionsLatest,
    InviteOptions as InviteOptionsLatest,
    UsageRequestOptions as UsageRequestOptionsLatest,
    UsageSummaryOptions as UsageSummaryOptionsLatest,
    UsageFieldsOptions as UsageFieldsOptionsLatest,
)

# responses
from .v1.response import (
    Message as MessageLatest,
    Project as ProjectLatest,
    ProjectsResponse as ProjectsResponseLatest,
    MembersResponse as MembersResponseLatest,
    Key as KeyLatest,
    KeyResponse as KeyResponseLatest,
    KeysResponse as KeysResponseLatest,
    ScopesResponse as ScopesResponseLatest,
    InvitesResponse as InvitesResponseLatest,
    UsageRequest as UsageRequestLatest,
    UsageRequestsResponse as UsageRequestsResponseLatest,
    UsageSummaryResponse as UsageSummaryResponseLatest,
    UsageFieldsResponse as UsageFieldsResponseLatest,
    Balance as BalanceLatest,
    BalancesResponse as BalancesResponseLatest,
)


"""
The client.py points to the current supported version in the SDK.
Older versions are supported in the SDK for backwards compatibility.
"""


# input
class ProjectOptions(ProjectOptionsLatest):
    """
    pass through for ProjectOptions based on API version
    """

    pass


class KeyOptions(KeyOptionsLatest):
    """
    pass through for KeyOptions based on API version
    """

    pass


class ScopeOptions(ScopeOptionsLatest):
    """
    pass through for ScopeOptions based on API version
    """

    pass


class InviteOptions(InviteOptionsLatest):
    """
    pass through for InviteOptions based on API version
    """

    pass


class UsageRequestOptions(UsageRequestOptionsLatest):
    """
    pass through for UsageRequestOptions based on API version
    """

    pass


class UsageSummaryOptions(UsageSummaryOptionsLatest):
    """
    pass through for UsageSummaryOptions based on API version
    """

    pass


class UsageFieldsOptions(UsageFieldsOptionsLatest):
    """
    pass through for UsageFieldsOptions based on API version
    """

    pass


# responses
class Message(MessageLatest):
    """
    pass through for Message based on API version
    """

    pass


class Project(ProjectLatest):
    """
    pass through for Project based on API version
    """

    pass


class ProjectsResponse(ProjectsResponseLatest):
    """
    pass through for ProjectsResponse based on API version
    """

    pass


class MembersResponse(MembersResponseLatest):
    """
    pass through for MembersResponse based on API version
    """

    pass


class Key(KeyLatest):
    """
    pass through for Key based on API version
    """

    pass


class KeyResponse(KeyResponseLatest):
    """
    pass through for KeyResponse based on API version
    """

    pass


class KeysResponse(KeysResponseLatest):
    """
    pass through for KeysResponse based on API version
    """

    pass


class ScopesResponse(ScopesResponseLatest):
    """
    pass through for ScopesResponse based on API version
    """

    pass


class InvitesResponse(InvitesResponseLatest):
    """
    pass through for InvitesResponse based on API version
    """

    pass


class UsageRequest(UsageRequestLatest):
    """
    pass through for UsageRequest based on API version
    """

    pass


class UsageRequestsResponse(UsageRequestsResponseLatest):
    """
    pass through for UsageRequestsResponse based on API version
    """

    pass


class UsageSummaryResponse(UsageSummaryResponseLatest):
    """
    pass through for UsageSummaryResponse based on API version
    """

    pass


class UsageFieldsResponse(UsageFieldsResponseLatest):
    """
    pass through for UsageFieldsResponse based on API version
    """

    pass


class Balance(BalanceLatest):
    """
    pass through for Balance based on API version
    """

    pass


class BalancesResponse(BalancesResponseLatest):
    """
    pass through for BalancesResponse based on API version
    """

    pass


# clients
class ManageClient(ManageClientLatest):
    """
    Please see ManageClientLatest for details
    """

    def __init__(self, config):
        super().__init__(config)


class AsyncManageClient(AsyncManageClientLatest):
    """
    Please see AsyncManageClientLatest for details
    """

    def __init__(self, config):
        super().__init__(config)



# Copyright 2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from dataclasses import dataclass, field
from dataclasses_json import dataclass_json, config

from io import BufferedReader
from typing import Union, Optional
from typing_extensions import TypedDict
import logging, verboselogs


@dataclass_json
@dataclass
class AnalyzeOptions:
    """
    Contains all the options for the AnalyzeOptions.

    Reference:
    https://developers.deepgram.com/reference/text-intelligence-apis
    """

    callback: Optional[str] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    callback_method: Optional[str] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    custom_intent: Optional[Union[list, str]] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    custom_intent_mode: Optional[str] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    custom_topic: Optional[Union[list, str]] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    custom_topic_mode: Optional[str] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    intents: Optional[bool] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    language: Optional[str] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    sentiment: Optional[bool] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    summarize: Optional[bool] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    topics: Optional[bool] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)

    def check(self):
        verboselogs.install()
        logger = logging.getLogger(__name__)
        logger.addHandler(logging.StreamHandler())
        prev = logger.level
        logger.setLevel(logging.ERROR)

        # no op at the moment

        logger.setLevel(prev)

        return True


class AnalyzeStreamSource(TypedDict):
    """
    Represents a data source for reading binary data from a stream-like source.

    This class is used to specify a source of binary data that can be read from
    a stream, such as an audio file in .wav format.

    Attributes:
        stream (BufferedReader): A BufferedReader object for reading binary data.
    """

    stream: BufferedReader


class UrlSource(TypedDict):
    """
    Represents a data source for specifying the location of a file via a URL.

    This class is used to specify a hosted file URL, typically pointing to an
    externally hosted file, such as an audio file hosted on a server or the internet.

    Attributes:
        url (str): The URL pointing to the hosted file.
    """

    url: str


class BufferSource(TypedDict):
    """
    Represents a data source for handling raw binary data.

    This class is used to specify raw binary data, such as audio data in its
    binary form, which can be captured from a microphone or generated synthetically.

    Attributes:
        buffer (bytes): The binary data.
    """

    buffer: bytes


AnalyzeSource = Union[UrlSource, BufferSource, AnalyzeStreamSource]
TextSource = Union[BufferSource, AnalyzeStreamSource]



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

import httpx
import json
import io
from typing import Dict, Optional, List

from .helpers import append_query_params
from ..options import DeepgramClientOptions
from .errors import DeepgramError, DeepgramApiError, DeepgramUnknownApiError


class AbstractSyncRestClient:
    """
    An abstract base class for a RESTful HTTP client.

    This class provides common HTTP methods (GET, POST, PUT, PATCH, DELETE) for making asynchronous HTTP requests.
    It handles error responses and provides basic JSON parsing.

    Args:
        url (Dict): The base URL for the RESTful API, including any path segments.
        headers (Optional[Dict[str, Any]]): Optional HTTP headers to include in requests.
        params (Optional[Dict[str, Any]]): Optional query parameters to include in requests.
        timeout (Optional[httpx.Timeout]): Optional timeout configuration for requests.

    Exceptions:
        DeepgramApiError: Raised for known API errors.
        DeepgramUnknownApiError: Raised for unknown API errors.
    """

    def __init__(self, config: DeepgramClientOptions):
        if config is None:
            raise DeepgramError("Config are required")
        self.config = config

    def get(
        self,
        url: str,
        options: Optional[Dict] = None,
        addons: Optional[Dict] = None,
        timeout: Optional[httpx.Timeout] = None,
        **kwargs,
    ) -> str:
        return self._handle_request(
            "GET",
            url,
            params=options,
            addons=addons,
            headers=self.config.headers,
            timeout=timeout,
            **kwargs,
        )

    def post_file(
        self,
        url: str,
        options: Optional[Dict] = None,
        addons: Optional[Dict] = None,
        timeout: Optional[httpx.Timeout] = None,
        file_result: Optional[List] = None,
        **kwargs,
    ) -> Dict:
        return self._handle_request(
            "POST",
            url,
            file_result=file_result,
            params=options,
            addons=addons,
            headers=self.config.headers,
            timeout=timeout,
            **kwargs,
        )

    def post(
        self,
        url: str,
        options: Optional[Dict] = None,
        addons: Optional[Dict] = None,
        timeout: Optional[httpx.Timeout] = None,
        **kwargs,
    ) -> str:
        return self._handle_request(
            "POST",
            url,
            params=options,
            addons=addons,
            headers=self.config.headers,
            timeout=timeout,
            **kwargs,
        )

    def put(
        self,
        url: str,
        options: Optional[Dict] = None,
        addons: Optional[Dict] = None,
        timeout: Optional[httpx.Timeout] = None,
        **kwargs,
    ) -> str:
        return self._handle_request(
            "PUT",
            url,
            params=options,
            addons=addons,
            headers=self.config.headers,
            timeout=timeout,
            **kwargs,
        )

    def patch(
        self,
        url: str,
        options: Optional[Dict] = None,
        addons: Optional[Dict] = None,
        timeout: Optional[httpx.Timeout] = None,
        **kwargs,
    ) -> str:
        return self._handle_request(
            "PATCH",
            url,
            params=options,
            addons=addons,
            headers=self.config.headers,
            timeout=timeout,
            **kwargs,
        )

    def delete(
        self,
        url: str,
        options: Optional[Dict] = None,
        addons: Optional[Dict] = None,
        timeout: Optional[httpx.Timeout] = None,
        **kwargs,
    ) -> str:
        return self._handle_request(
            "DELETE",
            url,
            params=options,
            addons=addons,
            headers=self.config.headers,
            timeout=timeout,
            **kwargs,
        )

    def _handle_request(
        self,
        method: str,
        url: str,
        params: Optional[Dict] = None,
        addons: Optional[Dict] = None,
        timeout: Optional[httpx.Timeout] = None,
        headers: Optional[Dict] = None,
        file_result: Optional[List] = None,
        **kwargs,
    ):
        new_url = url
        if params is not None:
            new_url = append_query_params(new_url, params)
        if addons is not None:
            new_url = append_query_params(new_url, addons)

        if timeout is None:
            timeout = httpx.Timeout(30.0, connect=10.0)

        try:
            with httpx.Client(timeout=timeout) as client:
                response = client.request(method, new_url, headers=headers, **kwargs)
                response.raise_for_status()

                # handle file response
                if file_result is not None:
                    ret = dict()
                    for item in file_result:
                        if item in response.headers:
                            ret[item] = response.headers[item]
                            continue
                        tmpItem = f"dg-{item}"
                        if tmpItem in response.headers:
                            ret[item] = response.headers[tmpItem]
                            continue
                        tmpItem = f"x-dg-{item}"
                        if tmpItem in response.headers:
                            ret[item] = response.headers[tmpItem]
                    ret["stream"] = io.BytesIO(response.content)
                    return ret

                # standard response
                return response.text

        except httpx._exceptions.HTTPError as e:
            if isinstance(e, httpx.HTTPStatusError):
                status_code = e.response.status_code or 500
                try:
                    json_object = json.loads(e.response.text)
                    raise DeepgramApiError(
                        json_object.get("err_msg"), status_code, json.dumps(json_object)
                    ) from e
                except json.decoder.JSONDecodeError:
                    raise DeepgramUnknownApiError(e.response.text, status_code) from e
                except ValueError as e:
                    raise DeepgramUnknownApiError(e.response.text, status_code) from e
            else:
                raise
        except Exception as e:
            raise



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from .client import PreRecordedClient
from .async_client import AsyncPreRecordedClient
from .options import (
    PrerecordedOptions,
    UrlSource,
    BufferSource,
    ReadStreamSource,
    FileSource,
    PrerecordedSource,
)
from ..enums import Sentiment
from .response import (
    AsyncPrerecordedResponse,
    PrerecordedResponse,
    SyncPrerecordedResponse,
)

from ....options import DeepgramClientOptions



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from enum import Enum

"""
Constants mapping to events from the Deepgram API
"""


class LiveTranscriptionEvents(str, Enum):
    Open: str = "Open"
    Close: str = "Close"
    Transcript: str = "Results"
    Metadata: str = "Metadata"
    UtteranceEnd: str = "UtteranceEnd"
    SpeechStarted: str = "SpeechStarted"
    Error: str = "Error"
    Warning: str = "Warning"



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from dataclasses import dataclass, field
from dataclasses_json import config, dataclass_json
from datetime import datetime
from typing import List, Optional, Dict

# Result Message


@dataclass_json
@dataclass
class OpenResponse:
    """
    Open Message from the Deepgram Platform
    """

    type: str = ""

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class Word:
    word: str = ""
    start: float = 0
    end: float = 0
    confidence: float = 0
    punctuated_word: Optional[str] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    speaker: Optional[int] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class Alternative:
    transcript: str = ""
    confidence: float = 0
    words: List[Word] = None

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "words" in _dict:
            _dict["words"] = [Word.from_dict(words) for words in _dict["words"]]
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class Channel:
    alternatives: List[Alternative] = None

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "alternatives" in _dict:
            _dict["alternatives"] = [
                Alternative.from_dict(alternatives)
                for alternatives in _dict["alternatives"]
            ]
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class ModelInfo:
    name: str = ""
    version: str = ""
    arch: str = ""

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class Metadata:
    request_id: str = ""
    model_info: ModelInfo = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    model_uuid: str = ""
    extra: Optional[Dict[str, str]] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "model_info" in _dict:
            _dict["model_info"] = [
                ModelInfo.from_dict(model_info) for model_info in _dict["model_info"]
            ]
        if "extra" in _dict:
            _dict["extra"] = [str(extra) for _, extra in _dict["extra"].items()]
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class LiveResultResponse:
    """
    Result Message from the Deepgram Platform
    """

    type: str = ""
    channel_index: List[int] = None
    duration: float = 0
    start: float = 0
    is_final: bool = False
    speech_final: bool = False
    channel: Channel = None
    metadata: Metadata = None

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "channel" in _dict:
            _dict["channel"] = [
                Channel.from_dict(channel) for channel in _dict["channel"]
            ]
        if "metadata" in _dict:
            _dict["metadata"] = [
                Metadata.from_dict(metadata) for metadata in _dict["metadata"]
            ]
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


# Metadata Message


@dataclass_json
@dataclass
class ModelInfo:
    name: str = ""
    version: str = ""
    arch: str = ""

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class MetadataResponse:
    """
    Metadata Message from the Deepgram Platform
    """

    type: str = ""
    transaction_key: str = ""
    request_id: str = ""
    sha256: str = ""
    created: str = ""
    duration: float = 0
    channels: int = 0
    models: List[str] = None
    model_info: Dict[str, ModelInfo] = None
    extra: Optional[Dict] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "models" in _dict:
            _dict["models"] = [str(models) for models in _dict["models"]]
        if "model_info" in _dict:
            _dict["model_info"] = [
                ModelInfo.from_dict(model_info)
                for _, model_info in _dict["model_info"].items()
            ]
        if "extra" in _dict:
            _dict["extra"] = [str(extra) for _, extra in _dict["extra"].items()]
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


# Speech Started Message


@dataclass_json
@dataclass
class SpeechStartedResponse:
    """
    SpeechStartedResponse Message from the Deepgram Platform
    """

    type: str = ""
    channel: List[int] = None
    timestamp: float = 0

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


# Utterance End Message


@dataclass_json
@dataclass
class UtteranceEndResponse:
    """
    UtteranceEnd Message from the Deepgram Platform
    """

    type: str = ""
    channel: List[int] = None
    last_word_end: float = 0

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


# Error Message


@dataclass_json
@dataclass
class ErrorResponse:
    """
    Error Message from the Deepgram Platform
    """

    description: str = ""
    message: str = ""
    type: str = ""
    variant: Optional[str] = ""

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class CloseResponse:
    """
    Close Message from the Deepgram Platform
    """

    type: str = ""

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)



# Copyright 2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from .client import SpeakClient
from .client import AsyncSpeakClient
from .client import SpeakOptions
from .client import SpeakResponse
from .client import (
    SpeakSource,
    TextSource,
    SpeakStreamSource,
)

from ...options import DeepgramClientOptions



# Copyright 2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from .client import SpeakClient
from .async_client import AsyncSpeakClient
from .options import SpeakOptions, SpeakSource, TextSource, SpeakStreamSource
from .response import SpeakResponse
from ....options import DeepgramClientOptions



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from .v1.client import OnPremClient as OnPremClientLatest
from .v1.async_client import AsyncOnPremClient as AsyncOnPremClientLatest

"""
The client.py points to the current supported version in the SDK.
Older versions are supported in the SDK for backwards compatibility.
"""


class OnPremClient(OnPremClientLatest):
    """
    Please see OnPremClientLatest for details
    """

    def __init__(self, config):
        super().__init__(config)


class AsyncOnPremClient(AsyncOnPremClientLatest):
    """
    Please see AsyncOnPremClientLatest for details
    """

    def __init__(self, config):
        super().__init__(config)



# Copyright 2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from dataclasses import dataclass, field
from dataclasses_json import config, dataclass_json
from typing import List, Optional, TypedDict, Dict
from ..enums import Sentiment


# Async Analyze Response Types:


@dataclass_json
@dataclass
class AsyncAnalyzeResponse:
    request_id: str = ""

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __str__(self) -> str:
        return self.to_json(indent=4)


# Analyze Response Types:


@dataclass_json
@dataclass
class IntentsInfo:
    model_uuid: str = ""
    input_tokens: int = 0
    output_tokens: int = 0

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class SentimentInfo:
    model_uuid: str = ""
    input_tokens: int = 0
    output_tokens: int = 0

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class SummaryInfo:
    model_uuid: str = ""
    input_tokens: int = 0
    output_tokens: int = 0

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class TopicsInfo:
    model_uuid: str = ""
    input_tokens: int = 0
    output_tokens: int = 0

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class Metadata:
    request_id: str = ""
    created: str = ""
    language: str = ""
    intents_info: Optional[IntentsInfo] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    sentiment_info: Optional[SentimentInfo] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    summary_info: Optional[SummaryInfo] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    topics_info: Optional[TopicsInfo] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "intents_info" in _dict:
            _dict["intents_info"] = IntentsInfo.from_dict(_dict["intents_info"])
        if "sentiment_info" in _dict:
            _dict["sentiment_info"] = SentimentInfo.from_dict(_dict["sentiment_info"])
        if "summary_info" in _dict:
            _dict["summary_info"] = SummaryInfo.from_dict(_dict["summary_info"])
        if "topics_info" in _dict:
            _dict["topics_info"] = TopicsInfo.from_dict(_dict["topics_info"])
        return _dict[key]

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class Average:
    sentiment: Sentiment = None
    sentiment_score: float = 0

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "sentiment" in _dict:
            _dict["sentiment"] = Sentiment.from_dict(_dict["sentiment"])
        return _dict[key]

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class Summary:
    text: str = ""

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class Topic:
    topic: str = ""
    confidence_score: float = 0

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class Intent:
    intent: str = ""
    confidence_score: float = 0

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class Segment:
    text: str = ""
    start_word: int = 0
    end_word: int = 0
    sentiment: Optional[Sentiment] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    sentiment_score: Optional[float] = 0
    intents: Optional[List[Intent]] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    topics: Optional[List[Topic]] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "sentiment" in _dict:
            _dict["sentiment"] = Sentiment.from_dict(_dict["sentiment"])
        if "intents" in _dict:
            _dict["intents"] = Intent.from_dict(_dict["intents"])
        if "topics" in _dict:
            _dict["topics"] = Topic.from_dict(_dict["topics"])
        return _dict[key]

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class Sentiments:
    segments: List[Segment] = None
    average: Average = None

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "segments" in _dict:
            _dict["segments"] = [
                Segment.from_dict(segments) for segments in _dict["segments"]
            ]
        if "average" in _dict:
            _dict["average"] = Average.from_dict(_dict["average"])
        return _dict[key]

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class Topics:
    segments: List[Segment] = None

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "segments" in _dict:
            _dict["segments"] = [
                Segment.from_dict(segments) for segments in _dict["segments"]
            ]
        return _dict[key]

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class Intents:
    segments: List[Segment] = None

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "segments" in _dict:
            _dict["segments"] = [
                Segment.from_dict(segments) for segments in _dict["segments"]
            ]
        return _dict[key]

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class Results:
    summary: Optional[Summary] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    sentiments: Optional[Sentiments] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    topics: Optional[Topics] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    intents: Optional[Intents] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "summary" in _dict:
            _dict["summary"] = Summary.from_dict(_dict["summary"])
        if "sentiments" in _dict:
            _dict["sentiments"] = Sentiments.from_dict(_dict["sentiments"])
        if "topics" in _dict:
            _dict["topics"] = Topics.from_dict(_dict["topics"])
        if "intents" in _dict:
            _dict["intents"] = Intents.from_dict(_dict["intents"])
        return _dict[key]

    def __str__(self) -> str:
        return self.to_json(indent=4)


# Analyze Response Result:


@dataclass_json
@dataclass
class AnalyzeResponse:
    metadata: Optional[Metadata] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    results: Optional[Results] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "metadata" in _dict:
            _dict["metadata"] = Metadata.from_dict(_dict["metadata"])
        if "results" in _dict:
            _dict["results"] = Results.from_dict(_dict["results"])
        return _dict[key]

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class SyncAnalyzeResponse(AnalyzeResponse):
    pass



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT
import json
from websockets.sync.client import connect
import websockets
import threading
import time
import logging, verboselogs
from typing import Dict, Union, Optional

from ....options import DeepgramClientOptions
from ..enums import LiveTranscriptionEvents
from ..helpers import convert_to_websocket_url, append_query_params
from ..errors import DeepgramError, DeepgramWebsocketError

from .response import (
    OpenResponse,
    LiveResultResponse,
    MetadataResponse,
    SpeechStartedResponse,
    UtteranceEndResponse,
    ErrorResponse,
    CloseResponse,
)
from .options import LiveOptions

ONE_SECOND = 1
DEEPGRAM_INTERVAL = 5
PING_INTERVAL = 20


class LiveClient:
    """
    Client for interacting with Deepgram's live transcription services over WebSockets.

     This class provides methods to establish a WebSocket connection for live transcription and handle real-time transcription events.

     Args:
         config (DeepgramClientOptions): all the options for the client.
    """

    def __init__(self, config: DeepgramClientOptions):
        if config is None:
            raise DeepgramError("Config are required")

        self.logger = logging.getLogger(__name__)
        self.logger.addHandler(logging.StreamHandler())
        self.logger.setLevel(config.verbose)

        self.config = config
        self.endpoint = "v1/listen"
        self._socket = None
        self.exit_event = None
        self._event_handlers = {event: [] for event in LiveTranscriptionEvents}
        self.websocket_url = convert_to_websocket_url(self.config.url, self.endpoint)

    # starts the WebSocket connection for live transcription
    def start(
        self,
        options: Optional[Union[LiveOptions, Dict]] = None,
        addons: Optional[Dict] = None,
        members: Optional[Dict] = None,
        **kwargs,
    ) -> bool:
        """
        Starts the WebSocket connection for live transcription.
        """
        self.logger.debug("LiveClient.start ENTER")
        self.logger.info("kwargs: %s", options)
        self.logger.info("addons: %s", addons)
        self.logger.info("members: %s", members)
        self.logger.info("options: %s", kwargs)

        if isinstance(options, LiveOptions) and not options.check():
            self.logger.error("options.check failed")
            self.logger.debug("LiveClient.start LEAVE")
            raise DeepgramError("Fatal transcription options error")

        if self._socket is not None:
            self.logger.error("socket is already initialized")
            self.logger.debug("LiveClient.start LEAVE")
            raise DeepgramWebsocketError("Websocket already started")

        self.options = options
        self.addons = addons

        # add "members" as members of the class
        if members is not None:
            self.__dict__.update(members)

        # set kwargs as members of the class
        if kwargs is not None:
            self.kwargs = kwargs
        else:
            self.kwargs = dict()

        if isinstance(options, LiveOptions):
            self.logger.info("LiveOptions switching class -> dict")
            self.options = self.options.to_dict()

        combined_options = self.options
        if addons is not None:
            self.logger.info("merging addons to options")
            combined_options.update(addons)
            self.logger.info("new options: %s", combined_options)
        self.logger.debug("combined_options: %s", combined_options)

        url_with_params = append_query_params(self.websocket_url, combined_options)
        self._socket = connect(url_with_params, additional_headers=self.config.headers)

        self.exit_event = threading.Event()
        self.lock_send = threading.Lock()

        # listening thread
        self.listening = threading.Thread(target=self._listening)
        self.listening.start()

        # keepalive thread
        self.processing = threading.Thread(target=self._keep_alive)
        self.processing.start()

        # push open event
        self._emit(
            LiveTranscriptionEvents.Open,
            OpenResponse(type=LiveTranscriptionEvents.Open.value),
        )

        self.logger.notice("start succeeded")
        self.logger.debug("LiveClient.start LEAVE")
        return True

    # registers event handlers for specific events
    def on(self, event, handler):
        """
        Registers event handlers for specific events.
        """
        self.logger.info("event fired: %s", event)
        if event in LiveTranscriptionEvents and callable(handler):
            self._event_handlers[event].append(handler)

    # unregisters event handlers for specific events
    def _emit(self, event, *args, **kwargs):
        for handler in self._event_handlers[event]:
            handler(self, *args, **kwargs)

    # main loop for handling incoming messages
    def _listening(self) -> None:
        self.logger.debug("LiveClient._listening ENTER")

        while True:
            try:
                message = self._socket.recv()

                if self.exit_event.is_set():
                    self.logger.notice("_listening exiting gracefully")
                    self.logger.debug("LiveClient._listening LEAVE")
                    return

                if len(message) == 0:
                    self.logger.info("message is empty")
                    continue

                data = json.loads(message)
                response_type = data.get("type")
                self.logger.debug("response_type: %s, data: %s", response_type, data)

                match response_type:
                    case LiveTranscriptionEvents.Open.value:
                        result = OpenResponse.from_json(message)
                        self.logger.verbose("OpenResponse: %s", result)
                        self._emit(
                            LiveTranscriptionEvents.Open,
                            open=result,
                            **dict(self.kwargs),
                        )
                    case LiveTranscriptionEvents.Transcript.value:
                        result = LiveResultResponse.from_json(message)
                        self.logger.verbose("LiveResultResponse: %s", result)
                        self._emit(
                            LiveTranscriptionEvents.Transcript,
                            result=result,
                            **dict(self.kwargs),
                        )
                    case LiveTranscriptionEvents.Metadata.value:
                        result = MetadataResponse.from_json(message)
                        self.logger.verbose("MetadataResponse: %s", result)
                        self._emit(
                            LiveTranscriptionEvents.Metadata,
                            metadata=result,
                            **dict(self.kwargs),
                        )
                    case LiveTranscriptionEvents.SpeechStarted.value:
                        result = SpeechStartedResponse.from_json(message)
                        self.logger.verbose("SpeechStartedResponse: %s", result)
                        self._emit(
                            LiveTranscriptionEvents.SpeechStarted,
                            speech_started=result,
                            **dict(self.kwargs),
                        )
                    case LiveTranscriptionEvents.UtteranceEnd.value:
                        result = UtteranceEndResponse.from_json(message)
                        self.logger.verbose("UtteranceEndResponse: %s", result)
                        self._emit(
                            LiveTranscriptionEvents.UtteranceEnd,
                            utterance_end=result,
                            **dict(self.kwargs),
                        )
                    case LiveTranscriptionEvents.Error.value:
                        result = ErrorResponse.from_json(message)
                        self.logger.verbose("ErrorResponse: %s", result)
                        self._emit(
                            LiveTranscriptionEvents.Error,
                            error=result,
                            **dict(self.kwargs),
                        )
                    case LiveTranscriptionEvents.Close.value:
                        result = CloseResponse.from_json(message)
                        self.logger.verbose("CloseResponse: %s", result)
                        self._emit(
                            LiveTranscriptionEvents.Close,
                            close=result,
                            **dict(self.kwargs),
                        )
                    case _:
                        self.logger.warning(
                            "Unknown Message: response_type: %s, data: %s",
                            response_type,
                            data,
                        )

            except websockets.exceptions.ConnectionClosedOK as e:
                self.logger.notice(f"_listening({e.code}) exiting gracefully")

                # signal exit and close
                self.signal_exit()

                self.logger.debug("LiveClient._listening LEAVE")
                return

            except websockets.exceptions.ConnectionClosedError as e:
                error: ErrorResponse = {
                    "type": "Exception",
                    "description": "ConnectionClosedError in _listening",
                    "message": f"{e}",
                    "variant": "",
                }
                self.logger.notice(
                    f"WebSocket connection in _listening closed with code {e.code}: {e.reason}"
                )
                self._emit(LiveTranscriptionEvents.Error, error)

                # signal exit and close
                self.signal_exit()

                self.logger.debug("LiveClient._listening LEAVE")

                if (
                    "termination_exception" in self.options
                    and self.options["termination_exception"] == "true"
                ):
                    raise
                return

            except Exception as e:
                error: ErrorResponse = {
                    "type": "Exception",
                    "description": "Exception in _listening",
                    "message": f"{e}",
                    "variant": "",
                }
                self.logger.error("Exception in _listening: %s", str(e))
                self._emit(LiveTranscriptionEvents.Error, error)

                # signal exit and close
                self.signal_exit()

                self.logger.debug("LiveClient._listening LEAVE")
                if (
                    "termination_exception" in self.options
                    and self.options["termination_exception"] == "true"
                ):
                    raise
                return

    def _keep_alive(self) -> None:
        self.logger.debug("LiveClient._keep_alive ENTER")

        counter = 0

        while True:
            try:
                counter += 1

                self.exit_event.wait(timeout=ONE_SECOND)
                if self.exit_event.is_set():
                    self.logger.notice("_keep_alive exiting gracefully")
                    self.logger.debug("LiveClient._keep_alive LEAVE")
                    return

                # deepgram keepalive
                if (
                    counter % DEEPGRAM_INTERVAL == 0
                    and self.config.options.get("keepalive") == "true"
                ):
                    self.logger.verbose("Sending KeepAlive...")
                    try:
                        self.send(json.dumps({"type": "KeepAlive"}))
                    except websockets.exceptions.WebSocketException as e:
                        self.logger.error("KeepAlive failed: %s", e)

                # websocket keepalive
                if counter % PING_INTERVAL == 0:
                    self.logger.verbose("Sending Protocol Ping...")
                    self.send_ping()

            except websockets.exceptions.ConnectionClosedOK as e:
                self.logger.notice(f"_keep_alive({e.code}) exiting gracefully")

                # signal exit and close
                self.signal_exit()

                self.logger.debug("LiveClient._keep_alive LEAVE")
                return

            except websockets.exceptions.ConnectionClosedError as e:
                error: ErrorResponse = {
                    "type": "Exception",
                    "description": "ConnectionClosedError in _keep_alive",
                    "message": f"{e}",
                    "variant": "",
                }
                self.logger.error(
                    f"WebSocket connection closed in _keep_alive with code {e.code}: {e.reason}"
                )
                self._emit(LiveTranscriptionEvents.Error, error)

                # signal exit and close
                self.signal_exit()

                self.logger.debug("LiveClient._keep_alive LEAVE")

                if (
                    "termination_exception" in self.options
                    and self.options["termination_exception"] == "true"
                ):
                    raise
                return

            except Exception as e:
                error: ErrorResponse = {
                    "type": "Exception",
                    "description": "Exception in _keep_alive",
                    "message": f"{e}",
                    "variant": "",
                }
                self._emit(LiveTranscriptionEvents.Error, error)
                self.logger.error("Exception in _keep_alive: %s", str(e))

                # signal exit and close
                self.signal_exit()

                self.logger.debug("LiveClient._keep_alive LEAVE")

                if (
                    "termination_exception" in self.options
                    and self.options["termination_exception"] == "true"
                ):
                    raise
                return

    # sends data over the WebSocket connection
    def send(self, data: Union[str, bytes]) -> int:
        """
        Sends data over the WebSocket connection.
        """
        self.logger.spam("LiveClient.send ENTER")
        self.logger.spam("data: %s", data)

        if self._socket:
            with self.lock_send:
                cnt = self._socket.send(data)

            self.logger.spam(f"send() succeeded. bytes: {cnt}")
            self.logger.spam("LiveClient.send LEAVE")
            return cnt

        self.logger.spam("send() failed. socket is empty")
        self.logger.spam("LiveClient.send LEAVE")
        return 0

    # sends a ping over the WebSocket connection
    def send_ping(self) -> None:
        """
        Sends a ping over the WebSocket connection.
        """
        self.logger.spam("LiveClient.send_ping ENTER")

        if self._socket:
            with self.lock_send:
                self.logger.debug("socket.ping() succeeded.")
                self._socket.ping()

        self.logger.spam("LiveClient.send_ping LEAVE")

    # closes the WebSocket connection gracefully
    def finish(self) -> bool:
        """
        Closes the WebSocket connection gracefully.
        """
        self.logger.spam("LiveClient.finish ENTER")

        if self._socket:
            self.logger.notice("sending CloseStream...")
            self.send(json.dumps({"type": "CloseStream"}))

            time.sleep(0.5)

            # push close event
            self._emit(
                LiveTranscriptionEvents.Close,
                CloseResponse(type=LiveTranscriptionEvents.Close.value),
            )

        # signal exit
        self.signal_exit()

        if self.processing is not None:
            self.processing.join()
            self.processing = None
        self.logger.notice("processing thread joined")

        if self.listening is not None:
            self.listening.join()
            self.listening = None
        self.logger.notice("listening thread joined")

        if self._socket is not None:
            self.logger.notice("closing socket...")
            self._socket.close()

        self._socket = None
        self.exit_event = None
        self.lock_send = None

        self.logger.notice("finish succeeded")
        self.logger.spam("LiveClient.finish LEAVE")
        return True

    # signals the WebSocket connection to exit
    def signal_exit(self) -> None:
        # signal exit
        self.exit_event.set()

        self.logger.notice("closing socket...")
        if self._socket:
            self.logger.debug("calling socket.close()")
            self._socket.close()
        self._socket = None



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from dataclasses import dataclass, field
from dataclasses_json import dataclass_json, config
from datetime import datetime
from typing import TypedDict, List, Optional


# Result Message


@dataclass_json
@dataclass
class Message:
    message: str = ""

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


# Projects


@dataclass_json
@dataclass
class Project:
    project_id: str = ""
    name: str = ""

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class ProjectsResponse:
    projects: List[Project] = None

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "projects" in _dict:
            _dict["projects"] = [
                Project.from_dict(projects) for projects in _dict["projects"]
            ]
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


# Members


@dataclass_json
@dataclass
class Member:
    email: str = ""
    first_name: str = ""
    last_name: str = ""
    member_id: str = ""

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class MembersResponse:
    members: List[Member] = None

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "members" in _dict:
            _dict["members"] = [
                Member.from_dict(members) for members in _dict["members"]
            ]
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


# Keys


@dataclass_json
@dataclass
class Key:
    api_key_id: str = ""
    comment: Optional[str] = ""
    created: str = ""
    scopes: List[str] = None

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "scopes" in _dict:
            _dict["scopes"] = [str(scopes) for scopes in _dict["scopes"]]
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class KeyResponse:
    api_key: Key = None
    member: Member = None

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "api_key" in _dict:
            _dict["api_key"] = Key.from_dict(_dict["api_key"])
        if "member" in _dict:
            _dict["member"] = Member.from_dict(_dict["member"])
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class KeysResponse:
    api_keys: List[KeyResponse] = None

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "api_keys" in _dict:
            _dict["api_keys"] = [
                KeyResponse.from_dict(api_keys) for api_keys in _dict["api_keys"]
            ]
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


# Scopes


@dataclass_json
@dataclass
class ScopesResponse:
    scopes: List[str] = None

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "scopes" in _dict:
            _dict["scopes"] = [str(scopes) for scopes in _dict["scopes"]]
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


# Invites


@dataclass_json
@dataclass
class Invite:
    email: str = ""
    scope: str = ""

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class InvitesResponse:
    invites: List[Invite] = None

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "invites" in _dict:
            _dict["invites"] = [
                Invite.from_dict(invites) for invites in _dict["invites"]
            ]
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


# Usage


@dataclass_json
@dataclass
class Config:
    language: str = ""
    model: str = ""
    punctuate: Optional[bool] = False
    utterances: Optional[bool] = False
    diarize: Optional[bool] = False
    smart_format: Optional[bool] = False
    interim_results: Optional[bool] = False

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class Details:
    usd: float = 0
    duration: float = 0
    total_audio: float = 0
    channels: int = 0
    streams: int = 0
    method: str = ""
    models: List[str] = None
    tags: Optional[List[str]] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    features: List[str] = None
    config: Config = None
    tier: Optional[str] = ""

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "models" in _dict:
            _dict["models"] = [str(models) for models in _dict["models"]]
        if "tags" in _dict:
            _dict["tags"] = [str(tags) for tags in _dict["tags"]]
        if "features" in _dict:
            _dict["features"] = [str(features) for features in _dict["features"]]
        if "config" in _dict:
            _dict["config"] = Config.from_dict(_dict["config"])
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class Callback:
    attempts: int = 0
    code: int = 0
    completed: str = ""

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class TokenDetail:
    feature: str = ""
    input: int = 0
    model: str = ""
    output: int = 0

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class Response:
    code: int = 0
    completed: str = ""
    details: Details = None
    token_details: List[TokenDetail] = None

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "details" in _dict:
            _dict["details"] = Details.from_dict(_dict["details"])
        if "token_details" in _dict:
            _dict["token_details"] = [
                TokenDetail.from_dict(token_details)
                for token_details in _dict["token_details"]
            ]
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class UsageRequest:
    project_uuid: str = ""
    request_id: str = ""
    created: str = ""
    path: str = ""
    api_key_id: str = ""
    response: Response = None
    callback: Optional[Callback] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    accessor: Optional[str] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "response" in _dict:
            _dict["response"] = Response.from_dict(_dict["response"])
        if "callback" in _dict:
            _dict["callback"] = Callback.from_dict(_dict["callback"])
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class UsageRequestsResponse:
    page: int = 0
    limit: int = 0
    requests: List[UsageRequest] = None

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "requests" in _dict:
            _dict["requests"] = [
                UsageRequest.from_dict(requests) for requests in _dict["requests"]
            ]
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class Tokens:
    tokens_in: int = 0
    out: int = 0

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class Results:
    start: str = ""
    end: str = ""
    hours: int = 0
    total_hours: int = 0
    requests: int = 0
    tokens: Tokens = None

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "tokens" in _dict:
            _dict["tokens"] = [Tokens.from_dict(tokens) for tokens in _dict["tokens"]]
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class Resolution:
    units: str = ""
    amount: int = 0

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class UsageSummaryResponse:
    start: str = ""
    end: str = ""
    resolution: Resolution = None
    results: List[Results] = None

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "resolution" in _dict:
            _dict["resolution"] = Resolution.from_dict(_dict["resolution"])
        if "results" in _dict:
            _dict["results"] = [
                Results.from_dict(results) for results in _dict["results"]
            ]
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class UsageModel:
    name: str = ""
    language: str = ""
    version: str = ""
    model_id: str = ""

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class UsageFieldsResponse:
    tags: Optional[List[str]] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    models: List[UsageModel] = None
    processing_methods: List[str] = None
    features: List[str] = None
    languages: List[str] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "tags" in _dict:
            _dict["tags"] = [str(tags) for tags in _dict["tags"]]
        if "models" in _dict:
            _dict["models"] = [
                UsageModel.from_dict(models) for models in _dict["models"]
            ]
        if "processing_methods" in _dict:
            _dict["processing_methods"] = [
                str(processing_methods)
                for processing_methods in _dict["processing_methods"]
            ]
        if "features" in _dict:
            _dict["features"] = [str(features) for features in _dict["features"]]
        if "languages" in _dict:
            _dict["languages"] = [str(languages) for languages in _dict["languages"]]
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


# Billing


@dataclass_json
@dataclass
class Balance:
    balance_id: str = ""
    amount: str = ""
    units: str = ""
    purchase_order_id: str = ""

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class BalancesResponse:
    balances: List[Balance] = None

    def __getitem__(self, key):
        _dict = self.to_dict()
        if "balances" in _dict:
            _dict["balances"] = [
                Balance.from_dict(balances) for balances in _dict["balances"]
            ]
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)



# Copyright 2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from .v1.client import SpeakClient as SpeakClientLatest
from .v1.async_client import AsyncSpeakClient as AsyncSpeakClientLatest
from .v1.options import (
    SpeakOptions as SpeakOptionsLatest,
    SpeakSource as SpeakSourceLatest,
    TextSource as TextSourceLatest,
    SpeakStreamSource as SpeakStreamSourceLatest,
)
from .v1.response import SpeakResponse as SpeakResponseLatest

"""
The client.py points to the current supported version in the SDK.
Older versions are supported in the SDK for backwards compatibility.
"""


# input
class SpeakOptions(SpeakOptionsLatest):
    """
    Please see SpeakOptionsLatest for details
    """

    pass


class TextSource(TextSourceLatest):
    """
    Please see TextSourceLatest for details
    """

    pass


class SpeakStreamSource(SpeakStreamSourceLatest):
    """
    Please see SpeakStreamSourceLatest for details
    """

    pass


SpeakSource = SpeakSourceLatest


# output
class SpeakResponse(SpeakResponseLatest):
    """
    Please see SpeakResponseLatest for details
    """

    pass


# clients
class SpeakClient(SpeakClientLatest):
    """
    Please see SpeakClientLatest for details
    """

    def __init__(self, config):
        self.config = config
        super().__init__(config)


class AsyncSpeakClient(AsyncSpeakClientLatest):
    """
    Please see AsyncSpeakClientLatest for details
    """

    def __init__(self, config):
        super().__init__(config)



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from .client import AnalyzeClient, AsyncAnalyzeClient
from .client import ReadClient, AsyncReadClient
from .client import AnalyzeOptions
from .client import (
    UrlSource,
    BufferSource,
    AnalyzeStreamSource,
    AnalyzeSource,
    TextSource,
)
from .client import Sentiment
from .client import AsyncAnalyzeResponse, AnalyzeResponse, SyncAnalyzeResponse

from ...options import DeepgramClientOptions



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from typing import Optional
from importlib import import_module
import logging, verboselogs
import os

# listen client
from .clients import Listen, Read

# live
from .clients import LiveClient, AsyncLiveClient
from .clients import (
    LiveOptions,
    LiveTranscriptionEvents,
)

# live client responses
from .clients import (
    OpenResponse,
    LiveResultResponse,
    MetadataResponse,
    SpeechStartedResponse,
    UtteranceEndResponse,
    ErrorResponse,
    CloseResponse,
)

# prerecorded
from .clients import PreRecordedClient, AsyncPreRecordedClient
from .clients import (
    FileSource,
    PrerecordedSource,
    UrlSource,
    BufferSource,
    ReadStreamSource,
    PrerecordedOptions,
    Sentiment,
)

# prerecorded client responses
from .clients import (
    AsyncPrerecordedResponse,
    PrerecordedResponse,
    SyncPrerecordedResponse,
)

# read
from .clients import ReadClient, AsyncReadClient
from .clients import AnalyzeClient, AsyncAnalyzeClient
from .clients import (
    AnalyzeSource,
    TextSource,
    UrlSource,
    BufferSource,
    AnalyzeStreamSource,
    AnalyzeOptions,
    Sentiment,
)

# read client responses
from .clients import (
    AsyncAnalyzeResponse,
    AnalyzeResponse,
    SyncAnalyzeResponse,
)

# speak client classes/input
from .clients import SpeakClient, AsyncSpeakClient
from .clients import SpeakOptions
from .clients import SpeakSource, TextSource, SpeakStreamSource

# speak client responses
from .clients import SpeakResponse

# manage client classes/input
from .clients import ManageClient, AsyncManageClient
from .clients import (
    ProjectOptions,
    KeyOptions,
    ScopeOptions,
    InviteOptions,
    UsageRequestOptions,
    UsageSummaryOptions,
    UsageFieldsOptions,
)

# manage client responses
from .clients import (
    Message,
    Project,
    ProjectsResponse,
    MembersResponse,
    Key,
    KeyResponse,
    KeysResponse,
    ScopesResponse,
    InvitesResponse,
    UsageRequest,
    UsageRequestsResponse,
    UsageSummaryResponse,
    UsageFieldsResponse,
    Balance,
    BalancesResponse,
)

# on-prem
from .clients.onprem.client import OnPremClient
from .clients.onprem.v1.async_client import AsyncOnPremClient

from .options import DeepgramClientOptions, ClientOptionsFromEnv
from .errors import DeepgramApiKeyError, DeepgramModuleError


class Deepgram:
    def __init__(self, *anything):
        raise Exception(
            """
            FATAL ERROR:
            You are attempting to instantiate a Deepgram object, which is no longer a class in version 3 of this SDK.

            To fix this issue:
                1. You need to revert to the previous version 2 of the SDK: pip install deepgram-sdk==2.12.0
                2. or, update your application's code to use version 3 of this SDK. See the README for more information.

            Things to consider:

                - This Version 3 of the SDK requires Python 3.10 or higher.
                  Older versions (3.9 and lower) of Python are nearing end-of-life: https://devguide.python.org/versions/
                  Understand the risks of using a version of Python nearing EOL.

                - Version 2 of the SDK will receive maintenance updates in the form of security fixes only.
                  No new features will be added to version 2 of the SDK.
            """
        )


class DeepgramClient:
    """
    Represents a client for interacting with the Deepgram API.

    This class provides a client for making requests to the Deepgram API with various configuration options.

    Attributes:
        api_key (str): The Deepgram API key used for authentication.
        config_options (DeepgramClientOptions): An optional configuration object specifying client options.

    Raises:
        DeepgramApiKeyError: If the API key is missing or invalid.

    Methods:
        listen: Returns a ListenClient instance for interacting with Deepgram's transcription services.

        manage: (Preferred) Returns a Threaded ManageClient instance for managing Deepgram resources.
        onprem: (Preferred) Returns an Threaded OnPremClient instance for interacting with Deepgram's on-premises API.

        asyncmanage: Returns an (Async) ManageClient instance for managing Deepgram resources.
        asynconprem: Returns an (Async) OnPremClient instance for interacting with Deepgram's on-premises API.
    """

    def __init__(
        self,
        api_key: str = "",
        config: Optional[DeepgramClientOptions] = None,
    ):
        verboselogs.install()
        self.logger = logging.getLogger(__name__)
        self.logger.addHandler(logging.StreamHandler())

        if api_key == "" and config is not None:
            self.logger.info("Attempting to set API key from config object")
            api_key = config.api_key
        if api_key == "":
            self.logger.info("Attempting to set API key from environment variable")
            api_key = os.getenv("DEEPGRAM_API_KEY", "")
        if api_key == "":
            self.logger.warning("WARNING: API key is missing")

        self.api_key = api_key
        if config is None:  # Use default configuration
            self.config = DeepgramClientOptions(self.api_key)
        else:
            config.set_apikey(self.api_key)
            self.config = config

    @property
    def listen(self):
        return Listen(self.config)

    @property
    def read(self):
        return Read(self.config)

    @property
    def speak(self):
        return self.Version(self.config, "speak")

    @property
    def asyncspeak(self):
        return self.Version(self.config, "asyncspeak")

    @property
    def manage(self):
        return self.Version(self.config, "manage")

    @property
    def asyncmanage(self):
        return self.Version(self.config, "asyncmanage")

    @property
    def onprem(self):
        return self.Version(self.config, "onprem")

    @property
    def asynconprem(self):
        return self.Version(self.config, "asynconprem")

    # INTERNAL CLASSES
    class Version:
        def __init__(self, config, parent: str):
            self.logger = logging.getLogger(__name__)
            self.logger.addHandler(logging.StreamHandler())
            self.logger.setLevel(config.verbose)
            self.config = config
            self.parent = parent

        # FUTURE VERSIONING:
        # When v2 or v1.1beta1 or etc. This allows easy access to the latest version of the API.
        # @property
        # def latest(self):
        #     match self.parent:
        #         case "manage":
        #             return ManageClient(self.config)
        #         case "onprem":
        #             return OnPremClient(self.config)
        #         case _:
        #             raise DeepgramModuleError("Invalid parent")

        def v(self, version: str = ""):
            self.logger.debug("Version.v ENTER")
            self.logger.info("version: %s", version)
            if len(version) == 0:
                self.logger.error("version is empty")
                self.logger.debug("Version.v LEAVE")
                raise DeepgramModuleError("Invalid module version")

            parent = ""
            fileName = ""
            className = ""
            match self.parent:
                case "manage":
                    parent = "manage"
                    fileName = "client"
                    className = "ManageClient"
                case "asyncmanage":
                    parent = "manage"
                    fileName = "async_client"
                    className = "AsyncManageClient"
                case "speak":
                    parent = "speak"
                    fileName = "client"
                    className = "SpeakClient"
                case "asyncspeak":
                    parent = "speak"
                    fileName = "async_client"
                    className = "AsyncSpeakClient"
                case "onprem":
                    parent = "onprem"
                    fileName = "client"
                    className = "OnPremClient"
                case "asynconprem":
                    parent = "onprem"
                    fileName = "async_client"
                    className = "AsyncOnPremClient"
                case _:
                    self.logger.error("parent unknown: %s", self.parent)
                    self.logger.debug("Version.v LEAVE")
                    raise DeepgramModuleError("Invalid parent type")

            # create class path
            path = f"deepgram.clients.{parent}.v{version}.{fileName}"
            self.logger.info("path: %s", path)
            self.logger.info("className: %s", className)

            # import class
            mod = import_module(path)
            if mod is None:
                self.logger.error("module path is None")
                self.logger.debug("Version.v LEAVE")
                raise DeepgramModuleError("Unable to find package")

            my_class = getattr(mod, className)
            if my_class is None:
                self.logger.error("my_class is None")
                self.logger.debug("Version.v LEAVE")
                raise DeepgramModuleError("Unable to find class")

            # instantiate class
            myClass = my_class(self.config)
            self.logger.notice("Version.v succeeded")
            self.logger.debug("Version.v LEAVE")
            return myClass



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

import logging, verboselogs
from typing import Dict, Optional
from .errors import DeepgramApiKeyError
import re
import os


class DeepgramClientOptions:
    """
    Represents options for configuring a Deepgram client.

    This class allows you to customize various options for interacting with the Deepgram API.

    Attributes:
        api_key: (Optional) A Deepgram API key used for authentication. Default uses the `DEEPGRAM_API_KEY` environment variable.
        verbose: (Optional) The logging level for the client. Defaults to `logging.WARNING`.
        url: (Optional) The URL used to interact with production, On-prem, and other Deepgram environments. Defaults to `api.deepgram.com`.
        headers: (Optional) Headers for initializing the client.
        options: (Optional) Additional options for initializing the client.
    """

    def __init__(
        self,
        api_key: str = "",
        url: str = "",
        verbose: int = logging.WARNING,
        headers: Optional[Dict] = None,
        options: Optional[Dict] = None,
    ):
        verboselogs.install()
        self.logger = logging.getLogger(__name__)
        self.logger.addHandler(logging.StreamHandler())

        self.verbose = verbose
        self.api_key = api_key
        self._update_headers(headers=headers)

        if len(url) == 0:
            url = "api.deepgram.com"
        self.url = self._get_url(url)

        if options is None:
            options = dict()
        self.options = options

    def set_apikey(self, api_key: str):
        self.api_key = api_key
        self._update_headers()

    def _get_url(self, url):
        if not re.match(r"^https?://", url, re.IGNORECASE):
            url = "https://" + url
        return url.strip("/")

    def _update_headers(self, headers: Optional[Dict] = None):
        if not hasattr(self, "headers") or self.headers is None:
            self.headers = {}
        self.headers["Accept"] = "application/json"
        if self.api_key:
            self.headers["Authorization"] = f"Token {self.api_key}"
        elif "Authorization" in self.headers:
            del self.headers["Authorization"]
        # Overwrite / add any headers that were passed in
        if headers:
            self.headers.update(headers)


class ClientOptionsFromEnv(DeepgramClientOptions):
    def __init__(
        self,
        api_key: str = "",
        url: str = "",
        verbose: int = logging.WARNING,
        headers: Optional[Dict] = None,
        options: Optional[Dict] = None,
    ):
        verboselogs.install()
        self.logger = logging.getLogger(__name__)
        self.logger.addHandler(logging.StreamHandler())
        self.logger.setLevel(logging.WARNING)  # temporary set for setup

        if api_key == "":
            api_key = os.getenv("DEEPGRAM_API_KEY", "")
            if api_key == "":
                self.logger.critical("Deepgram API KEY is not set")
                raise DeepgramApiKeyError("Deepgram API KEY is not set")

        if url == "":
            url = os.getenv("DEEPGRAM_HOST", "api.deepgram.com")
            self.logger.notice(f"Deepgram host is set to {url}")

        if verbose == logging.WARNING:
            _loglevel = os.getenv("DEEPGRAM_LOGGING", "")
            if _loglevel != "":
                verbose = int(_loglevel)
            if type(verbose) != int:
                match verbose:
                    case "NOTSET":
                        self.logger.notice("Logging level is set to NOTSET")
                        verbose = logging.NOTSET
                    case "SPAM":
                        self.logger.notice("Logging level is set to SPAM")
                        verbose = logging.SPAM
                    case "DEBUG":
                        self.logger.notice("Logging level is set to DEBUG")
                        verbose = logging.DEBUG
                    case "VERBOSE":
                        self.logger.notice("Logging level is set to VERBOSE")
                        verbose = logging.VERBOSE
                    case "NOTICE":
                        self.logger.notice("Logging level is set to NOTICE")
                        verbose = logging.NOTICE
                    case "WARNING":
                        self.logger.notice("Logging level is set to WARNING")
                        verbose = logging.WARNING
                    case "SUCCESS":
                        self.logger.notice("Logging level is set to SUCCESS")
                        verbose = logging.SUCCESS
                    case "ERROR":
                        self.logger.notice("Logging level is set to ERROR")
                        verbose = logging.ERROR
                    case "CRITICAL":
                        self.logger.notice("Logging level is set to CRITICAL")
                        verbose = logging.CRITICAL
                    case _:
                        self.logger.notice("Logging level is set to WARNING")
                        verbose = logging.WARNING
        self.logger.notice(f"Logging level is set to {verbose}")

        if headers is None:
            headers = dict()
            for x in range(0, 20):
                header = os.getenv(f"DEEPGRAM_HEADER_{x}", None)
                if header is not None:
                    headers[header] = os.getenv(f"DEEPGRAM_HEADER_VALUE_{x}", None)
                    self.logger.debug(
                        f"Deepgram header {header} is set with value {headers[header]}"
                    )
                else:
                    break
            if len(headers) == 0:
                self.logger.notice("Deepgram headers are not set")
                headers = None

        if options is None:
            options = dict()
            for x in range(0, 20):
                param = os.getenv(f"DEEPGRAM_PARAM_{x}", None)
                if param is not None:
                    options[param] = os.getenv(f"DEEPGRAM_PARAM_VALUE_{x}", None)
                    self.logger.debug(
                        f"Deepgram option {param} is set with value {options[param]}"
                    )
                else:
                    break
            if len(options) == 0:
                self.logger.notice("Deepgram options are not set")
                options = None

        super().__init__(
            api_key=api_key, url=url, verbose=verbose, headers=headers, options=options
        )



# Copyright 2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from importlib import import_module
import logging, verboselogs
from typing import Union

from ..options import DeepgramClientOptions
from .errors import DeepgramModuleError

# live client
# classes and input
from .analyze import (
    AnalyzeClient,
    AsyncAnalyzeClient,
    AnalyzeOptions,
)

# responses
from .analyze import (
    AsyncAnalyzeResponse,
    AnalyzeResponse,
    SyncAnalyzeResponse,
)


class Read:
    """
    Represents a client for interacting with the Deepgram API.

    This class provides a client for making requests to the Deepgram API with various configuration options.

    Attributes:
        api_key (str): The Deepgram API key used for authentication.
        config_options (DeepgramClientOptions): An optional configuration object specifying client options.

    Raises:
        DeepgramApiKeyError: If the API key is missing or invalid.

    Methods:
        read: (Preferred) Returns an Threaded AnalyzeClient instance for interacting with Deepgram's read transcription services.
        asyncread: Returns an (Async) AnalyzeClient instance for interacting with Deepgram's read transcription services.
    """

    def __init__(self, config: DeepgramClientOptions):
        self.logger = logging.getLogger(__name__)
        self.logger.addHandler(logging.StreamHandler())
        self.logger.setLevel(config.verbose)
        self.config = config

    @property
    def analyze(self):
        return self.Version(self.config, "analyze")

    @property
    def asyncanalyze(self):
        return self.Version(self.config, "asyncanalyze")

    # INTERNAL CLASSES
    class Version:
        def __init__(self, config, parent: str):
            self.logger = logging.getLogger(__name__)
            self.logger.addHandler(logging.StreamHandler())
            self.logger.setLevel(config.verbose)
            self.config = config
            self.parent = parent

        # FUTURE VERSIONING:
        # When v2 or v1.1beta1 or etc. This allows easy access to the latest version of the API.
        # @property
        # def latest(self):
        #     match self.parent:
        #         case "analyze":
        #             return AnalyzeClient(self.config)
        #         case _:
        #             raise DeepgramModuleError("Invalid parent")

        def v(self, version: str = ""):
            self.logger.debug("Version.v ENTER")
            self.logger.info("version: %s", version)
            if len(version) == 0:
                self.logger.error("version is empty")
                self.logger.debug("Version.v LEAVE")
                raise DeepgramModuleError("Invalid module version")

            parent = ""
            fileName = ""
            className = ""
            match self.parent:
                case "analyze":
                    parent = "analyze"
                    fileName = "client"
                    className = "AnalyzeClient"
                case "asyncanalyze":
                    parent = "analyze"
                    fileName = "async_client"
                    className = "AsyncAnalyzeClient"
                case _:
                    self.logger.error("parent unknown: %s", self.parent)
                    self.logger.debug("Version.v LEAVE")
                    raise DeepgramModuleError("Invalid parent type")

            # create class path
            path = f"deepgram.clients.{parent}.v{version}.{fileName}"
            self.logger.info("path: %s", path)
            self.logger.info("className: %s", className)

            # import class
            mod = import_module(path)
            if mod is None:
                self.logger.error("module path is None")
                self.logger.debug("Version.v LEAVE")
                raise DeepgramModuleError("Unable to find package")

            my_class = getattr(mod, className)
            if my_class is None:
                self.logger.error("my_class is None")
                self.logger.debug("Version.v LEAVE")
                raise DeepgramModuleError("Unable to find class")

            # instantiate class
            myClass = my_class(self.config)
            self.logger.notice("Version.v succeeded")
            self.logger.debug("Version.v LEAVE")
            return myClass



# Copyright 2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from typing import Union

from .v1.client import AnalyzeClient as AnalyzeClientLatest
from .v1.async_client import AsyncAnalyzeClient as AsyncAnalyzeClientLatest
from .v1.options import (
    AnalyzeOptions as AnalyzeOptionsLatest,
    AnalyzeStreamSource as AnalyzeStreamSourceLatest,
    BufferSource as BufferSourceLatest,
    UrlSource as UrlSourceLatest,
    AnalyzeSource as AnalyzeSourceLatest,
    TextSource as TextSourceLatest,
)
from .v1.response import (
    SyncAnalyzeResponse as SyncAnalyzeResponseLatest,
    AnalyzeResponse as AnalyzeResponseLatest,
    AsyncAnalyzeResponse as AsyncAnalyzeResponseLatest,
)
from .enums import Sentiment

"""
The client.py points to the current supported version in the SDK.
Older versions are supported in the SDK for backwards compatibility.
"""


# input
class AnalyzeOptions(AnalyzeOptionsLatest):
    """
    Please see AnalyzeOptionsLatest for details
    """

    pass


class AnalyzeStreamSource(AnalyzeStreamSourceLatest):
    """
    Please see AnalyzeStreamSourceLatest for details
    """

    pass


class BufferSource(BufferSourceLatest):
    """
    Please see BufferSourceLatest for details
    """

    pass


class UrlSource(UrlSourceLatest):
    """
    Please see UrlSourceLatest for details
    """

    pass


AnalyzeSource = AnalyzeSourceLatest
TextSource = TextSourceLatest


# responses


class AsyncAnalyzeResponse(AsyncAnalyzeResponseLatest):
    """
    Please see AnalyzeResponseLatest for details
    """

    pass


class AnalyzeResponse(AnalyzeResponseLatest):
    """
    Please see AnalyzeResponseLatest for details
    """

    pass


class SyncAnalyzeResponse(SyncAnalyzeResponseLatest):
    """
    Please see AnalyzeResponseLatest for details
    """

    pass


# clients
class AnalyzeClient(AnalyzeClientLatest):
    """
    Please see AnalyzeClientLatest for details
    """

    def __init__(self, config):
        super().__init__(config)


class AsyncAnalyzeClient(AsyncAnalyzeClientLatest):
    """
    Please see AsyncAnalyzeClientLatest for details
    """

    def __init__(self, config):
        super().__init__(config)


class ReadClient(AnalyzeClientLatest):
    """
    Just an alias for the Analyze Client
    """

    def __init__(self, config):
        self.config = config
        super().__init__(config)


class AsyncReadClient(AsyncAnalyzeClientLatest):
    """
    Just an alias for the Async Analyze Client
    """

    def __init__(self, config):
        super().__init__(config)



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from dataclasses import dataclass, field
from dataclasses_json import dataclass_json, config
from datetime import datetime
from typing import TypedDict, List, Optional

# Input


@dataclass_json
@dataclass
class ProjectOptions:
    name: str = ""

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class KeyOptions:
    comment: Optional[str] = ""
    time_to_live_in_seconds: Optional[int] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    expiration_date: Optional[str] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    scopes: List[str] = None
    tags: Optional[List[str]] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )

    def __getitem__(self, key):
        _dict = self.to_dict()
        if _dict["scopes"] is not None:
            _dict["scopes"] = [str(scopes) for scopes in _dict["scopes"]]
        if _dict["tags"] is not None:
            _dict["tags"] = [str(tags) for tags in _dict["tags"]]
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class ScopeOptions:
    scope: str = ""

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class InviteOptions:
    email: str = ""
    scope: str = ""

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class UsageRequestOptions:
    start: Optional[str] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    end: Optional[str] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    limit: Optional[int] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    status: Optional[str] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class UsageSummaryOptions:
    start: Optional[str] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    end: Optional[str] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    accessor: Optional[str] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    tag: Optional[str] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    method: Optional[str] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    model: Optional[str] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    multichannel: Optional[bool] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    interim_results: Optional[bool] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    punctuate: Optional[bool] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    ner: Optional[bool] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    utterances: Optional[bool] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    replace: Optional[bool] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    profanity_filter: Optional[bool] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    keywords: Optional[bool] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    detect_topics: Optional[bool] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    diarize: Optional[bool] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    search: Optional[bool] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    redact: Optional[bool] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    alternatives: Optional[bool] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    numerals: Optional[bool] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    smart_format: Optional[bool] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)


@dataclass_json
@dataclass
class UsageFieldsOptions:
    start: Optional[str] = ""
    end: Optional[str] = ""

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)



# Copyright 2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from .client import AnalyzeClient
from .async_client import AsyncAnalyzeClient
from .options import (
    AnalyzeOptions,
    UrlSource,
    BufferSource,
    AnalyzeStreamSource,
    TextSource,
    AnalyzeSource,
)
from ....options import DeepgramClientOptions
from .response import AsyncAnalyzeResponse, AnalyzeResponse
from ..enums import Sentiment



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

# version
__version__ = "v3.2.0-dev.6"

# entry point for the deepgram python sdk
from .client import Deepgram, DeepgramClient
from .options import DeepgramClientOptions, ClientOptionsFromEnv
import logging, verboselogs

# listen client
from .client import Listen, Read

# live
from .client import LiveTranscriptionEvents
from .client import LiveClient, AsyncLiveClient
from .client import LiveOptions
from .client import (
    OpenResponse,
    LiveResultResponse,
    MetadataResponse,
    SpeechStartedResponse,
    UtteranceEndResponse,
    ErrorResponse,
    CloseResponse,
)

# prerecorded
from .client import PreRecordedClient, AsyncPreRecordedClient
from .client import (
    FileSource,
    PrerecordedSource,
    UrlSource,
    BufferSource,
    ReadStreamSource,
    PrerecordedOptions,
    Sentiment,
)
from .client import (
    AsyncPrerecordedResponse,
    PrerecordedResponse,
    SyncPrerecordedResponse,
)

# read
from .client import ReadClient, AsyncReadClient
from .client import AnalyzeClient, AsyncAnalyzeClient
from .client import (
    AnalyzeSource,
    TextSource,
    UrlSource,
    BufferSource,
    AnalyzeStreamSource,
    AnalyzeOptions,
    Sentiment,
)
from .client import (
    AsyncAnalyzeResponse,
    AnalyzeResponse,
    SyncAnalyzeResponse,
)

# speak
from .client import SpeakClient, AsyncSpeakClient
from .client import SpeakSource, TextSource, SpeakStreamSource, SpeakOptions
from .client import SpeakResponse

# manage
from .client import ManageClient, AsyncManageClient
from .client import (
    ProjectOptions,
    KeyOptions,
    ScopeOptions,
    InviteOptions,
    UsageRequestOptions,
    UsageSummaryOptions,
    UsageFieldsOptions,
)

# manage client responses
from .client import (
    Message,
    Project,
    ProjectsResponse,
    MembersResponse,
    Key,
    KeyResponse,
    KeysResponse,
    ScopesResponse,
    InvitesResponse,
    UsageRequest,
    UsageRequestsResponse,
    UsageSummaryResponse,
    UsageFieldsResponse,
    Balance,
    BalancesResponse,
)

# onprem
from .client import (
    OnPremClient,
    AsyncOnPremClient,
)

# utilities
from .audio import Microphone
from .audio import (
    LOGGING,
    CHANNELS,
    RATE,
    CHUNK,
)



# Copyright 2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

import httpx
import logging, verboselogs
import json
from typing import Dict, Union, Optional

from ...abstract_sync_client import AbstractSyncRestClient
from ..errors import DeepgramError, DeepgramTypeError
from .helpers import is_text_source

from .options import SpeakOptions, SpeakSource, TextSource, SpeakStreamSource
from .response import SpeakResponse


class SpeakClient(AbstractSyncRestClient):
    """
    A client class for doing Text-to-Speech.
    Provides methods for speaking from text.
    """

    def __init__(self, config):
        self.logger = logging.getLogger(__name__)
        self.logger.addHandler(logging.StreamHandler())
        self.logger.setLevel(config.verbose)
        self.config = config
        super().__init__(config)

    """
    Speak from a text source.

    Args:
        source (TextSource): The text source to speak.
        options (SpeakOptions): Additional options for the ingest (default is None).
        endpoint (str): The API endpoint for the ingest (default is "v1/speak").

    Returns:
        SpeakResponse: The response from the speak request.

    Raises:
        DeepgramTypeError: Raised for known API errors.
    """

    def stream(
        self,
        source: SpeakSource,
        options: Optional[Union[Dict, SpeakOptions]] = None,
        addons: Optional[Dict] = None,
        timeout: Optional[httpx.Timeout] = None,
        endpoint: str = "v1/speak",
    ) -> SpeakResponse:
        self.logger.debug("SpeakClient.stream ENTER")

        url = f"{self.config.url}/{endpoint}"
        if is_text_source(source):
            body = source
        else:
            self.logger.error("Unknown speak source type")
            self.logger.debug("SpeakClient.stream LEAVE")
            raise DeepgramTypeError("Unknown speak source type")

        if isinstance(options, SpeakOptions) and not options.check():
            self.logger.error("options.check failed")
            self.logger.debug("SpeakClient.stream LEAVE")
            raise DeepgramError("Fatal speak options error")

        self.logger.info("url: %s", url)
        self.logger.info("source: %s", source)
        if isinstance(options, SpeakOptions):
            self.logger.info("SpeakOptions switching class -> json")
            options = json.loads(options.to_json())
        self.logger.info("options: %s", options)
        self.logger.info("addons: %s", addons)

        returnVals = [
            "content-type",
            "request-id",
            "model-uuid",
            "model-name",
            "char-count",
            "transfer-encoding",
            "date",
        ]
        result = self.post_file(
            url,
            options=options,
            addons=addons,
            json=body,
            timeout=timeout,
            file_result=returnVals,
        )
        self.logger.info("result: %s", result)
        sResp = SpeakResponse(
            content_type=result["content-type"],
            request_id=result["request-id"],
            model_uuid=result["model-uuid"],
            model_name=result["model-name"],
            characters=result["char-count"],
            transfer_encoding=result["transfer-encoding"],
            date=result["date"],
            stream=result["stream"],
        )
        self.logger.verbose("result: %s", sResp)
        self.logger.notice("speak succeeded")
        self.logger.debug("SpeakClient.stream LEAVE")
        return sResp

    def save(
        self,
        filename: str,
        source: SpeakSource,
        options: Optional[Union[Dict, SpeakOptions]] = None,
        addons: Optional[Dict] = None,
        timeout: Optional[httpx.Timeout] = None,
        endpoint: str = "v1/speak",
    ) -> SpeakResponse:
        self.logger.debug("SpeakClient.save ENTER")

        res = self.stream(source, options, addons, timeout, endpoint)

        # save to file
        file = open(filename, "wb+")
        file.write(res.stream.getbuffer())
        file.flush()
        file.close()

        # add filename to response
        res.stream = None
        res.filename = filename

        self.logger.debug("SpeakClient.save LEAVE")
        return res



# Copyright 2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from dataclasses import dataclass, field
from dataclasses_json import config, dataclass_json
from typing import List, Optional, TypedDict, Dict

import io

# Speak Response Types:


@dataclass_json
@dataclass
class SpeakResponse:
    content_type: str = ""
    request_id: str = ""
    model_uuid: str = ""
    model_name: str = ""
    characters: str = 0
    transfer_encoding: str = ""
    date: str = ""
    stream: Optional[io.BytesIO] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    filename: Optional[str] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    # this is a hack to make the response look like a dict because of the io.BytesIO object
    # otherwise it will throw an exception on printing
    def __str__(self) -> str:
        myDict = self.to_dict()
        return myDict.__str__()



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from .v1.client import LiveClient as LiveClientLatest
from .v1.async_client import AsyncLiveClient as AsyncLiveClientLatest
from .v1.options import LiveOptions as LiveOptionsLatest
from .enums import LiveTranscriptionEvents
from .v1.response import (
    OpenResponse as OpenResponseLatest,
    LiveResultResponse as LiveResultResponseLatest,
    MetadataResponse as MetadataResponseLatest,
    SpeechStartedResponse as SpeechStartedResponseLatest,
    UtteranceEndResponse as UtteranceEndResponseLatest,
    ErrorResponse as ErrorResponseLatest,
    CloseResponse as CloseResponseLatest,
)

"""
The vX/client.py points to the current supported version in the SDK.
Older versions are supported in the SDK for backwards compatibility.
"""


# input
class LiveOptions(LiveOptionsLatest):
    """
    pass through for LiveOptions based on API version
    """

    pass


# responses
class OpenResponse(OpenResponseLatest):
    """
    pass through for OpenResponse based on API version
    """

    pass


class LiveResultResponse(LiveResultResponseLatest):
    """
    pass through for LiveResultResponse based on API version
    """

    pass


class MetadataResponse(MetadataResponseLatest):
    """
    pass through for MetadataResponse based on API version
    """

    pass


class SpeechStartedResponse(SpeechStartedResponseLatest):
    """
    pass through for SpeechStartedResponse based on API version
    """

    pass


class UtteranceEndResponse(UtteranceEndResponseLatest):
    """
    pass through for UtteranceEndResponse based on API version
    """

    pass


class ErrorResponse(ErrorResponseLatest):
    """
    pass through for ErrorResponse based on API version
    """

    pass


class CloseResponse(CloseResponseLatest):
    """
    pass through for CloseResponse based on API version
    """

    pass


# clients
class LiveClient(LiveClientLatest):
    """
    Please see LiveClientLatest for details
    """

    def __init__(self, config):
        super().__init__(config)


class AsyncLiveClient(AsyncLiveClientLatest):
    """
    Please see LiveClientLatest for details
    """

    def __init__(self, config):
        super().__init__(config)



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from .client import ManageClient
from .async_client import AsyncManageClient
from .options import (
    ProjectOptions,
    KeyOptions,
    ScopeOptions,
    InviteOptions,
    UsageRequestOptions,
    UsageSummaryOptions,
    UsageFieldsOptions,
)
from .response import (
    Message,
    Project,
    ProjectsResponse,
    MembersResponse,
    Key,
    KeyResponse,
    KeysResponse,
    ScopesResponse,
    InvitesResponse,
    UsageRequest,
    UsageRequestsResponse,
    UsageSummaryResponse,
    UsageFieldsResponse,
    Balance,
    BalancesResponse,
)
from ....options import DeepgramClientOptions



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

import logging, verboselogs

from ...abstract_async_client import AbstractAsyncRestClient


class AsyncOnPremClient(AbstractAsyncRestClient):
    """
    Client for interacting with Deepgram's on-premises API.

    This class provides methods to manage and interact with on-premises projects and distribution credentials.

    Args:
        config (DeepgramClientOptions): all the options for the client.
    """

    def __init__(self, config):
        self.logger = logging.getLogger(__name__)
        self.logger.addHandler(logging.StreamHandler())
        self.logger.setLevel(config.verbose)
        self.config = config
        self.endpoint = "v1/projects"
        super().__init__(config)

    async def list_onprem_credentials(self, project_id: str):
        self.logger.debug("OnPremClient.list_onprem_credentials ENTER")
        url = f"{self.config.url}/{self.endpoint}/{project_id}/onprem/distribution/credentials"
        self.logger.info("url: %s", url)
        self.logger.info("project_id: %s", project_id)
        res = await self.get(url)
        self.logger.verbose("result: %s", res)
        self.logger.notice("list_onprem_credentials succeeded")
        self.logger.debug("OnPremClient.list_onprem_credentials LEAVE")
        return res

    async def get_onprem_credentials(
        self, project_id: str, distribution_credentials_id: str
    ):
        self.logger.debug("OnPremClient.get_onprem_credentials ENTER")
        url = f"{self.config.url}/{self.endpoint}/{project_id}/onprem/distribution/credentials/{distribution_credentials_id}"
        self.logger.info("url: %s", url)
        self.logger.info("project_id: %s", project_id)
        self.logger.info("distribution_credentials_id: %s", distribution_credentials_id)
        res = await self.get(url)
        self.logger.verbose("result: %s", res)
        self.logger.notice("get_onprem_credentials succeeded")
        self.logger.debug("OnPremClient.get_onprem_credentials LEAVE")
        return res

    async def create_onprem_credentials(self, project_id: str, options):
        self.logger.debug("OnPremClient.create_onprem_credentials ENTER")
        url = f"{self.config.url}/{self.endpoint}/{project_id}/onprem/distribution/credentials/"
        self.logger.info("url: %s", url)
        self.logger.info("project_id: %s", project_id)
        self.logger.info("options: %s", options)
        res = await self.post(url, json=options)
        self.logger.verbose("result: %s", res)
        self.logger.notice("create_onprem_credentials succeeded")
        self.logger.debug("OnPremClient.create_onprem_credentials LEAVE")
        return res

    async def delete_onprem_credentials(
        self, project_id: str, distribution_credentials_id: str
    ):
        self.logger.debug("OnPremClient.delete_onprem_credentials ENTER")
        url = f"{self.config.url}/{self.endpoint}/{project_id}/onprem/distribution/credentials/{distribution_credentials_id}"
        self.logger.info("url: %s", url)
        self.logger.info("project_id: %s", project_id)
        self.logger.info("distrbution_credentials_id: %s", distribution_credentials_id)
        res = await self.delete(url)
        self.logger.verbose("result: %s", res)
        self.logger.notice("delete_onprem_credentials succeeded")
        self.logger.debug("OnPremClient.delete_onprem_credentials LEAVE")
        return res



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from .v1.client import PreRecordedClient as PreRecordedClientLatest
from .v1.async_client import AsyncPreRecordedClient as AsyncPreRecordedClientLatest
from .v1.options import (
    PrerecordedOptions as PrerecordedOptionsLatest,
    UrlSource as UrlSourceLatest,
    BufferSource as BufferSourceLatest,
    ReadStreamSource as ReadStreamSourceLatest,
    FileSource as FileSourceLatest,
    PrerecordedSource as PrerecordedSourceLatest,
)
from .v1.response import (
    AsyncPrerecordedResponse as AsyncPrerecordedResponseLatest,
    PrerecordedResponse as PrerecordedResponseLatest,
    SyncPrerecordedResponse as SyncPrerecordedResponseLatest,
)
from .enums import Sentiment

"""
The client.py points to the current supported version in the SDK.
Older versions are supported in the SDK for backwards compatibility.
"""


# input
class PrerecordedOptions(PrerecordedOptionsLatest):
    """
    Please see PrerecordedOptionsLatest for details
    """

    pass


class UrlSource(UrlSourceLatest):
    """
    Please see UrlSourceLatest for details
    """

    pass


class BufferSource(BufferSourceLatest):
    """
    Please see BufferSourceLatest for details
    """

    pass


class ReadStreamSource(ReadStreamSourceLatest):
    """
    Please see ReadStreamSourceLatest for details
    """

    pass


FileSource = FileSourceLatest
PrerecordedSource = PrerecordedSourceLatest


# output
class AsyncPrerecordedResponse(AsyncPrerecordedResponseLatest):
    """
    Please see AsyncPrerecordedResponseLatest for details
    """

    pass


class PrerecordedResponse(PrerecordedResponseLatest):
    """
    Please see PrerecordedResponseLatest for details
    """

    pass


class SyncPrerecordedResponse(PrerecordedResponseLatest):
    """
    Please see PrerecordedResponseLatest for details
    """

    pass


# clients
class PreRecordedClient(PreRecordedClientLatest):
    """
    Please see PreRecordedClientLatest for details
    """

    def __init__(self, config):
        self.config = config
        super().__init__(config)


class AsyncPreRecordedClient(AsyncPreRecordedClientLatest):
    """
    Please see AsyncPreRecordedClientLatest for details
    """

    def __init__(self, config):
        super().__init__(config)



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

# listen
from .listen import Listen
from .read import Read

# live
from .live import LiveClient, AsyncLiveClient
from .live import LiveOptions
from .live import LiveTranscriptionEvents
from .live import (
    OpenResponse,
    LiveResultResponse,
    MetadataResponse,
    SpeechStartedResponse,
    UtteranceEndResponse,
    ErrorResponse,
    CloseResponse,
)

# prerecorded
from .prerecorded import PreRecordedClient, AsyncPreRecordedClient
from .prerecorded import PrerecordedOptions
from .prerecorded import Sentiment
from .prerecorded import (
    FileSource,
    PrerecordedSource,
    UrlSource,
    BufferSource,
    ReadStreamSource,
)
from .prerecorded import (
    AsyncPrerecordedResponse,
    PrerecordedResponse,
    SyncPrerecordedResponse,
)

# read
from .analyze import ReadClient, AsyncReadClient
from .analyze import AnalyzeClient, AsyncAnalyzeClient
from .analyze import AnalyzeOptions
from .analyze import Sentiment
from .analyze import (
    AnalyzeSource,
    TextSource,
    UrlSource,
    BufferSource,
    AnalyzeStreamSource,
)
from .analyze import (
    AsyncAnalyzeResponse,
    AnalyzeResponse,
    SyncAnalyzeResponse,
)

# speak
from .speak import SpeakClient, AsyncSpeakClient
from .speak import SpeakOptions
from .speak import (
    SpeakSource,
    TextSource,
    SpeakStreamSource,
)
from .speak import SpeakResponse

# manage
from .manage import ManageClient, AsyncManageClient
from .manage import (
    ProjectOptions,
    KeyOptions,
    ScopeOptions,
    InviteOptions,
    UsageRequestOptions,
    UsageSummaryOptions,
    UsageFieldsOptions,
)
from .manage import (
    Message,
    Project,
    ProjectsResponse,
    MembersResponse,
    Key,
    KeyResponse,
    KeysResponse,
    ScopesResponse,
    InvitesResponse,
    UsageRequest,
    UsageRequestsResponse,
    UsageSummaryResponse,
    UsageFieldsResponse,
    Balance,
    BalancesResponse,
)

# onprem
from .onprem import OnPremClient, AsyncOnPremClient

# client options
from ..options import DeepgramClientOptions



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

import httpx
import logging, verboselogs
import json
from typing import Dict, Union, Optional

from ...abstract_sync_client import AbstractSyncRestClient
from ..errors import DeepgramError, DeepgramTypeError
from .helpers import is_buffer_source, is_readstream_source, is_url_source
from ..enums import Sentiment

from .options import (
    PrerecordedOptions,
    UrlSource,
    BufferSource,
    ReadStreamSource,
    FileSource,
    PrerecordedSource,
)
from .response import AsyncPrerecordedResponse, PrerecordedResponse


class PreRecordedClient(AbstractSyncRestClient):
    """
    A client class for handling pre-recorded audio data.
    Provides methods for transcribing audio from URLs and files.
    """

    def __init__(self, config):
        self.logger = logging.getLogger(__name__)
        self.logger.addHandler(logging.StreamHandler())
        self.logger.setLevel(config.verbose)
        self.config = config
        super().__init__(config)

    """
    Transcribes audio from a URL source.

    Args:
        source (UrlSource): The URL source of the audio to transcribe.
        options (PrerecordedOptions): Additional options for the transcription (default is None).
        endpoint (str): The API endpoint for the transcription (default is "v1/listen").

    Returns:
        PrerecordedResponse: An object containing the transcription result.

    Raises:
        DeepgramTypeError: Raised for known API errors.
    """

    def transcribe_url(
        self,
        source: UrlSource,
        options: Optional[Union[Dict, PrerecordedOptions]] = None,
        addons: Optional[Dict] = None,
        timeout: Optional[httpx.Timeout] = None,
        endpoint: str = "v1/listen",
    ) -> PrerecordedResponse:
        self.logger.debug("PreRecordedClient.transcribe_url ENTER")

        if (options is Dict and "callback" in options is not None) or (
            isinstance(options, PrerecordedOptions) and options.callback is not None
        ):
            self.logger.debug("PreRecordedClient.transcribe_url LEAVE")
            return self.transcribe_url_callback(
                source, options.callback, options, addons, timeout, endpoint
            )

        url = f"{self.config.url}/{endpoint}"
        if is_url_source(source):
            body = source
        else:
            self.logger.error("Unknown transcription source type")
            self.logger.debug("PreRecordedClient.transcribe_url LEAVE")
            raise DeepgramTypeError("Unknown transcription source type")

        if isinstance(options, PrerecordedOptions) and not options.check():
            self.logger.error("options.check failed")
            self.logger.debug("PreRecordedClient.transcribe_url LEAVE")
            raise DeepgramError("Fatal transcription options error")

        self.logger.info("url: %s", url)
        self.logger.info("source: %s", source)
        if isinstance(options, PrerecordedOptions):
            self.logger.info("PrerecordedOptions switching class -> json")
            options = json.loads(options.to_json())
        self.logger.info("options: %s", options)
        self.logger.info("addons: %s", addons)
        result = self.post(
            url, options=options, addons=addons, json=body, timeout=timeout
        )
        self.logger.info("json: %s", result)
        res = PrerecordedResponse.from_json(result)
        self.logger.verbose("result: %s", res)
        self.logger.notice("transcribe_url succeeded")
        self.logger.debug("PreRecordedClient.transcribe_url LEAVE")
        return res

    """
    Transcribes audio from a URL source and sends the result to a callback URL.

    Args:
        source (UrlSource): The URL source of the audio to transcribe.
        callback (str): The callback URL where the transcription results will be sent.
        options (PrerecordedOptions): Additional options for the transcription (default is None).
        endpoint (str): The API endpoint for the transcription (default is "v1/listen").

    Returns:
        AsyncPrerecordedResponse: An object containing the request_id or an error message.

    Raises:
        DeepgramTypeError: Raised for known API errors.
    """

    def transcribe_url_callback(
        self,
        source: UrlSource,
        callback: str,
        options: Optional[Union[Dict, PrerecordedOptions]] = None,
        addons: Optional[Dict] = None,
        timeout: Optional[httpx.Timeout] = None,
        endpoint: str = "v1/listen",
    ) -> AsyncPrerecordedResponse:
        self.logger.debug("PreRecordedClient.transcribe_url_callback ENTER")

        url = f"{self.config.url}/{endpoint}"
        if options is None:
            options = {}
        if isinstance(options, PrerecordedOptions):
            options.callback = callback
        else:
            options["callback"] = callback
        if is_url_source(source):
            body = source
        else:
            self.logger.error("Unknown transcription source type")
            self.logger.debug("PreRecordedClient.transcribe_url_callback LEAVE")
            raise DeepgramTypeError("Unknown transcription source type")

        if isinstance(options, PrerecordedOptions) and not options.check():
            self.logger.error("options.check failed")
            self.logger.debug("PreRecordedClient.transcribe_url_callback LEAVE")
            raise DeepgramError("Fatal transcription options error")

        self.logger.info("url: %s", url)
        self.logger.info("source: %s", source)
        if isinstance(options, PrerecordedOptions):
            self.logger.info("PrerecordedOptions switching class -> json")
            options = json.loads(options.to_json())
        self.logger.info("options: %s", options)
        self.logger.info("addons: %s", addons)
        result = self.post(
            url, options=options, addons=addons, json=body, timeout=timeout
        )
        self.logger.info("json: %s", result)
        res = AsyncPrerecordedResponse.from_json(result)
        self.logger.verbose("result: %s", res)
        self.logger.notice("transcribe_url_callback succeeded")
        self.logger.debug("PreRecordedClient.transcribe_url_callback LEAVE")
        return res

    """
    Transcribes audio from a local file source.

    Args:
        source (FileSource): The local file source of the audio to transcribe.
        options (PrerecordedOptions): Additional options for the transcription (default is None).
        endpoint (str): The API endpoint for the transcription (default is "v1/listen").

    Returns:
        PrerecordedResponse: An object containing the transcription result or an error message.

    Raises:
        DeepgramTypeError: Raised for known API errors.
    """

    def transcribe_file(
        self,
        source: FileSource,
        options: Optional[Union[Dict, PrerecordedOptions]] = None,
        addons: Optional[Dict] = None,
        timeout: Optional[httpx.Timeout] = None,
        endpoint: str = "v1/listen",
    ) -> PrerecordedResponse:
        self.logger.debug("PreRecordedClient.transcribe_file ENTER")

        if (options is Dict and "callback" in options is not None) or (
            isinstance(options, PrerecordedOptions) and options.callback is not None
        ):
            self.logger.debug("PreRecordedClient.transcribe_file LEAVE")
            return self.transcribe_file_callback(
                source, options.callback, options, addons, timeout, endpoint
            )

        url = f"{self.config.url}/{endpoint}"
        if is_buffer_source(source):
            body = source["buffer"]
        elif is_readstream_source(source):
            body = source["stream"]
        else:
            self.logger.error("Unknown transcription source type")
            self.logger.debug("PreRecordedClient.transcribe_file LEAVE")
            raise DeepgramTypeError("Unknown transcription source type")

        if isinstance(options, PrerecordedOptions) and not options.check():
            self.logger.error("options.check failed")
            self.logger.debug("PreRecordedClient.transcribe_file LEAVE")
            raise DeepgramError("Fatal transcription options error")

        self.logger.info("url: %s", url)
        if isinstance(options, PrerecordedOptions):
            self.logger.info("PrerecordedOptions switching class -> json")
            options = json.loads(options.to_json())
        self.logger.info("options: %s", options)
        self.logger.info("addons: %s", addons)
        result = self.post(
            url, options=options, addons=addons, content=body, timeout=timeout
        )
        self.logger.info("json: %s", result)
        res = PrerecordedResponse.from_json(result)
        self.logger.verbose("result: %s", res)
        self.logger.notice("transcribe_file succeeded")
        self.logger.debug("PreRecordedClient.transcribe_file LEAVE")
        return res

    """
    Transcribes audio from a local file source and sends the result to a callback URL.

    Args:
        source (FileSource): The local file source of the audio to transcribe.
        callback (str): The callback URL where the transcription results will be sent.
        options (PrerecordedOptions): Additional options for the transcription (default is None).
        endpoint (str): The API endpoint for the transcription (default is "v1/listen").

    Returns:
        AsyncPrerecordedResponse: An object containing the request_id or an error message.

    Raises:
        DeepgramTypeError: Raised for known API errors.
    """

    def transcribe_file_callback(
        self,
        source: FileSource,
        callback: str,
        options: Optional[Union[Dict, PrerecordedOptions]] = None,
        addons: Optional[Dict] = None,
        timeout: Optional[httpx.Timeout] = None,
        endpoint: str = "v1/listen",
    ) -> AsyncPrerecordedResponse:
        self.logger.debug("PreRecordedClient.transcribe_file_callback ENTER")

        url = f"{self.config.url}/{endpoint}"
        if options is None:
            options = {}
        if isinstance(options, PrerecordedOptions):
            options.callback = callback
        else:
            options["callback"] = callback
        if is_buffer_source(source):
            body = source["buffer"]
        elif is_readstream_source(source):
            body = source["stream"]
        else:
            self.logger.error("Unknown transcription source type")
            self.logger.debug("PreRecordedClient.transcribe_file_callback LEAVE")
            raise DeepgramTypeError("Unknown transcription source type")

        if isinstance(options, PrerecordedOptions) and not options.check():
            self.logger.error("options.check failed")
            self.logger.debug("PreRecordedClient.transcribe_file_callback LEAVE")
            raise DeepgramError("Fatal transcription options error")

        self.logger.info("url: %s", url)
        if isinstance(options, PrerecordedOptions):
            self.logger.info("PrerecordedOptions switching class -> json")
            options = json.loads(options.to_json())
        self.logger.info("options: %s", options)
        self.logger.info("addons: %s", addons)
        result = self.post(
            url, options=options, addons=addons, content=body, timeout=timeout
        )
        self.logger.info("json: %s", result)
        res = AsyncPrerecordedResponse.from_json(result)
        self.logger.verbose("result: %s", res)
        self.logger.notice("transcribe_file_callback succeeded")
        self.logger.debug("PreRecordedClient.transcribe_file_callback LEAVE")
        return res



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from dataclasses import dataclass, field
from dataclasses_json import dataclass_json, config
from typing import List, Optional, Union
import logging, verboselogs


@dataclass_json
@dataclass
class LiveOptions:
    """
    Live Transcription Options for the Deepgram Platform.

    Please see the documentation for more information on each option:
    https://developers.deepgram.com/reference/streaming
    """

    alternatives: Optional[int] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    callback: Optional[str] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    callback_method: Optional[str] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    channels: Optional[int] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    diarize: Optional[bool] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    diarize_version: Optional[str] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    encoding: Optional[str] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    endpointing: Optional[str] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    extra: Optional[Union[list, str]] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    filler_words: Optional[bool] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    interim_results: Optional[bool] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    keywords: Optional[str] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    language: Optional[str] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    model: Optional[str] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    multichannel: Optional[bool] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    numerals: Optional[bool] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    punctuate: Optional[bool] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    profanity_filter: Optional[bool] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    redact: Optional[bool] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    replace: Optional[str] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    sample_rate: Optional[int] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    search: Optional[str] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    smart_format: Optional[bool] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    tag: Optional[list] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    tier: Optional[str] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    utterance_end_ms: Optional[str] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    vad_events: Optional[bool] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    version: Optional[str] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)

    def check(self):
        verboselogs.install()
        logger = logging.getLogger(__name__)
        logger.addHandler(logging.StreamHandler())
        prev = logger.level
        logger.setLevel(logging.ERROR)

        if self.tier:
            logger.error(
                "WARNING: Tier is deprecated. Will be removed in a future version."
            )

        logger.setLevel(prev)

        return True



# Copyright 2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from dataclasses import dataclass, field
from dataclasses_json import dataclass_json, config

from io import BufferedReader
from typing import Union, List, TypedDict, Optional
import logging, verboselogs


@dataclass_json
@dataclass
class SpeakOptions:
    """
    Contains all the options for the SpeakOptions.

    Reference:
    https://developers.deepgram.com/reference/text-to-speech-preview-api
    """

    model: Optional[str] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    encoding: Optional[str] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    container: Optional[str] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    sample_rate: Optional[int] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )
    bit_rate: Optional[int] = field(
        default=None, metadata=config(exclude=lambda f: f is None)
    )

    def __getitem__(self, key):
        _dict = self.to_dict()
        return _dict[key]

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __str__(self) -> str:
        return self.to_json(indent=4)

    def check(self):
        verboselogs.install()
        logger = logging.getLogger(__name__)
        logger.addHandler(logging.StreamHandler())
        prev = logger.level
        logger.setLevel(logging.ERROR)

        # no op at the moment

        logger.setLevel(prev)

        return True


class SpeakStreamSource(TypedDict):
    """
    Represents a data source for reading binary data from a stream-like source.

    This class is used to specify a source of binary data that can be read from
    a stream, such as an audio file in .wav format.

    Attributes:
        stream (BufferedReader): A BufferedReader object for reading binary data.
    """

    stream: BufferedReader


class TextSource(TypedDict):
    """
    Represents a data source for reading binary data from a text-like source.

    This class is used to specify a source of text data that can be read from.

    Attributes:
        text (str): A string for reading text data.
    """

    text: str


SpeakSource = Union[TextSource, SpeakStreamSource]



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT
import asyncio
import json
import websockets
import logging, verboselogs
from typing import Dict, Union, Optional

from ....options import DeepgramClientOptions
from ..enums import LiveTranscriptionEvents
from ..helpers import convert_to_websocket_url, append_query_params
from ..errors import DeepgramError, DeepgramWebsocketError

from .response import (
    OpenResponse,
    LiveResultResponse,
    MetadataResponse,
    SpeechStartedResponse,
    UtteranceEndResponse,
    ErrorResponse,
    CloseResponse,
)
from .options import LiveOptions

ONE_SECOND = 1
DEEPGRAM_INTERVAL = 5
PING_INTERVAL = 20


class AsyncLiveClient:
    """
    Client for interacting with Deepgram's live transcription services over WebSockets.

     This class provides methods to establish a WebSocket connection for live transcription and handle real-time transcription events.

     Args:
         config (DeepgramClientOptions): all the options for the client.
    """

    def __init__(self, config: DeepgramClientOptions):
        if config is None:
            raise DeepgramError("Config are required")

        self.logger = logging.getLogger(__name__)
        self.logger.addHandler(logging.StreamHandler())
        self.logger.setLevel(config.verbose)

        self.config = config
        self.endpoint = "v1/listen"
        self._socket = None
        self._event_handlers = {event: [] for event in LiveTranscriptionEvents}
        self.websocket_url = convert_to_websocket_url(self.config.url, self.endpoint)

    # starts the WebSocket connection for live transcription
    async def start(
        self,
        options: Optional[Union[LiveOptions, Dict]] = None,
        addons: Optional[Dict] = None,
        members: Optional[Dict] = None,
        **kwargs,
    ) -> bool:
        self.logger.debug("AsyncLiveClient.start ENTER")
        self.logger.info("kwargs: %s", options)
        self.logger.info("addons: %s", addons)
        self.logger.info("members: %s", members)
        self.logger.info("options: %s", kwargs)

        if isinstance(options, LiveOptions) and not options.check():
            self.logger.error("options.check failed")
            self.logger.debug("AsyncLiveClient.start LEAVE")
            raise DeepgramError("Fatal transcription options error")

        if self._socket is not None:
            self.logger.error("socket is already initialized")
            self.logger.debug("AsyncLiveClient.start LEAVE")
            raise DeepgramWebsocketError("Websocket already started")

        self.options = options
        self.addons = addons

        # add "members" as members of the class
        if members is not None:
            self.__dict__.update(members)

        # add "kwargs" as members of the class
        if kwargs is not None:
            self.kwargs = kwargs
        else:
            self.kwargs = dict()

        if isinstance(options, LiveOptions):
            self.logger.info("LiveOptions switching class -> dict")
            self.options = self.options.to_dict()

        combined_options = self.options
        if addons is not None:
            self.logger.info("merging addons to options")
            combined_options.update(addons)
            self.logger.info("new options: %s", combined_options)
        self.logger.debug("combined_options: %s", combined_options)

        url_with_params = append_query_params(self.websocket_url, combined_options)
        try:
            self._socket = await _socket_connect(url_with_params, self.config.headers)

            self._listen_thread = asyncio.create_task(self._listening())
            self._keep_alive_thread = asyncio.create_task(self._keep_alive())

            # push open event
            await self._emit(
                LiveTranscriptionEvents.Open,
                OpenResponse(type=LiveTranscriptionEvents.Open.value),
            )

            self.logger.notice("start succeeded")
            self.logger.debug("AsyncLiveClient.start LEAVE")
            return True
        except websockets.ConnectionClosed as e:
            await self._emit(LiveTranscriptionEvents.Close, e.code)
            self.logger.notice("exception: websockets.ConnectionClosed")
            self.logger.debug("AsyncLiveClient.start LEAVE")

    # registers event handlers for specific events
    def on(self, event: LiveTranscriptionEvents, handler) -> None:
        if event in LiveTranscriptionEvents and callable(handler):
            self._event_handlers[event].append(handler)

    # triggers the registered event handlers for a specific event
    async def _emit(self, event: LiveTranscriptionEvents, *args, **kwargs) -> None:
        for handler in self._event_handlers[event]:
            asyncio.create_task(handler(self, *args, **kwargs))

    # main loop for handling incoming messages
    async def _listening(self) -> None:
        self.logger.debug("AsyncLiveClient._listening ENTER")

        try:
            async for message in self._socket:
                data = json.loads(message)
                response_type = data.get("type")
                self.logger.debug("response_type: %s, data: %s", response_type, data)

                match response_type:
                    case LiveTranscriptionEvents.Open.value:
                        result = OpenResponse.from_json(message)
                        self.logger.verbose("OpenResponse: %s", result)
                        await self._emit(
                            LiveTranscriptionEvents.Open,
                            open=result,
                            **dict(self.kwargs),
                        )
                    case LiveTranscriptionEvents.Transcript.value:
                        result = LiveResultResponse.from_json(message)
                        self.logger.verbose("LiveResultResponse: %s", result)
                        await self._emit(
                            LiveTranscriptionEvents.Transcript,
                            result=result,
                            **dict(self.kwargs),
                        )
                    case LiveTranscriptionEvents.Metadata.value:
                        result = MetadataResponse.from_json(message)
                        self.logger.verbose("MetadataResponse: %s", result)
                        await self._emit(
                            LiveTranscriptionEvents.Metadata,
                            metadata=result,
                            **dict(self.kwargs),
                        )
                    case LiveTranscriptionEvents.SpeechStarted.value:
                        result = SpeechStartedResponse.from_json(message)
                        self.logger.verbose("SpeechStartedResponse: %s", result)
                        await self._emit(
                            LiveTranscriptionEvents.SpeechStarted,
                            speech_started=result,
                            **dict(self.kwargs),
                        )
                    case LiveTranscriptionEvents.UtteranceEnd.value:
                        result = UtteranceEndResponse.from_json(message)
                        self.logger.verbose("UtteranceEndResponse: %s", result)
                        await self._emit(
                            LiveTranscriptionEvents.UtteranceEnd,
                            utterance_end=result,
                            **dict(self.kwargs),
                        )
                    case LiveTranscriptionEvents.Error.value:
                        result = ErrorResponse.from_json(message)
                        self.logger.verbose("LiveTranscriptionEvents: %s", result)
                        await self._emit(
                            LiveTranscriptionEvents.Error,
                            error=result,
                            **dict(self.kwargs),
                        )
                    case LiveTranscriptionEvents.Close.value:
                        result = CloseResponse.from_json(message)
                        self.logger.verbose("CloseResponse: %s", result)
                        await self._emit(
                            LiveTranscriptionEvents.Close,
                            close=result,
                            **dict(self.kwargs),
                        )
                    case _:
                        error = ErrorResponse(
                            type="UnhandledMessage",
                            description="Unknown message type",
                            message=f"Unhandle message type: {response_type}",
                        )
                        await self._emit(LiveTranscriptionEvents.Error, error=error)

        except websockets.exceptions.ConnectionClosedOK as e:
            self.logger.notice(f"_listening({e.code}) exiting gracefully")
            self.logger.debug("AsyncLiveClient._listening LEAVE")
            return

        except websockets.exceptions.WebSocketException as e:
            error: ErrorResponse = {
                "type": "Exception",
                "description": "WebSocketException in _listening",
                "message": f"{e}",
                "variant": "",
            }
            self.logger.notice(
                f"WebSocket exception in _listening with code {e.code}: {e.reason}"
            )
            await self._emit(LiveTranscriptionEvents.Error, error)

            self.logger.debug("AsyncLiveClient._listening LEAVE")

            if (
                "termination_exception" in self.options
                and self.options["termination_exception"] == "true"
            ):
                raise

        except Exception as e:
            error: ErrorResponse = {
                "type": "Exception",
                "description": "Exception in _listening",
                "message": f"{e}",
                "variant": "",
            }
            self.logger.error("Exception in _listening: %s", str(e))
            await self._emit(LiveTranscriptionEvents.Error, error)

            self.logger.debug("AsyncLiveClient._listening LEAVE")

            if (
                "termination_exception" in self.options
                and self.options["termination_exception"] == "true"
            ):
                raise

    # keep the connection alive by sending keepalive messages
    async def _keep_alive(self) -> None:
        self.logger.debug("AsyncLiveClient._keep_alive ENTER")

        counter = 0
        while True:
            try:
                counter += 1
                await asyncio.sleep(ONE_SECOND)

                if self._socket is None:
                    self.logger.notice("socket is None, exiting keep_alive")
                    self.logger.debug("AsyncLiveClient._keep_alive LEAVE")
                    break

                # deepgram keepalive
                if (
                    counter % DEEPGRAM_INTERVAL == 0
                    and self.config.options.get("keepalive") == "true"
                ):
                    self.logger.verbose("Sending KeepAlive...")
                    try:
                        await self.send(json.dumps({"type": "KeepAlive"}))
                    except websockets.exceptions.WebSocketException as e:
                        self.logger.error("KeepAlive failed: %s", e)

            except websockets.exceptions.ConnectionClosedOK as e:
                self.logger.notice(f"_keep_alive({e.code}) exiting gracefully")
                self.logger.debug("AsyncLiveClient._keep_alive LEAVE")
                return

            except websockets.exceptions.ConnectionClosedError as e:
                error: ErrorResponse = {
                    "type": "Exception",
                    "description": "ConnectionClosedError in _keep_alive",
                    "message": f"{e}",
                    "variant": "",
                }
                self.logger.error(
                    f"WebSocket connection closed in _keep_alive with code {e.code}: {e.reason}"
                )
                await self._emit(LiveTranscriptionEvents.Error, error)

                self.logger.debug("AsyncLiveClient._keep_alive LEAVE")

                if (
                    "termination_exception" in self.options
                    and self.options["termination_exception"] == "true"
                ):
                    raise
                return

            except Exception as e:
                error: ErrorResponse = {
                    "type": "Exception",
                    "description": "Exception in _keep_alive",
                    "message": f"{e}",
                    "variant": "",
                }
                await self._emit(LiveTranscriptionEvents.Error, error)
                self.logger.error("Exception in _keep_alive: %s", str(e))

                self.logger.debug("AsyncLiveClient._keep_alive LEAVE")

                if (
                    "termination_exception" in self.options
                    and self.options["termination_exception"] == "true"
                ):
                    raise
                return

        self.logger.debug("AsyncLiveClient._keep_alive LEAVE")

    # sends data over the WebSocket connection
    async def send(self, data: Union[str, bytes]) -> int:
        """
        Sends data over the WebSocket connection.
        """
        self.logger.spam("AsyncLiveClient.send ENTER")
        self.logger.spam("data: %s", data)

        if self._socket is not None:
            cnt = await self._socket.send(data)
            self.logger.spam(f"send() succeeded. bytes: {cnt}")
            self.logger.spam("AsyncLiveClient.send LEAVE")
            return cnt

        self.logger.error("send() failed. socket is None")
        self.logger.spam("AsyncLiveClient.send LEAVE")
        return 0

    async def finish(self) -> bool:
        """
        Closes the WebSocket connection gracefully.
        """
        self.logger.debug("AsyncLiveClient.finish ENTER")

        if self._socket:
            self.logger.notice("send CloseStream...")
            await self._socket.send(json.dumps({"type": "CloseStream"}))

            await asyncio.sleep(0.5)

            # push close event
            await self._emit(
                LiveTranscriptionEvents.Close,
                CloseResponse(type=LiveTranscriptionEvents.Close.value),
            )

            self.logger.notice("socket.wait_closed...")
            try:
                await self._socket.wait_closed()
            except websockets.exceptions.WebSocketException as e:
                self.logger.error("socket.wait_closed failed: %s", e)
            self.logger.notice("socket.wait_closed succeeded")

        self.logger.notice("cancelling tasks...")
        try:
            # Before cancelling, check if the tasks were created
            if self._listen_thread is not None:
                self._listen_thread.cancel()
            if self._keep_alive_thread is not None:
                self._keep_alive_thread.cancel()

            # Use asyncio.gather to wait for tasks to be cancelled
            tasks = [self._listen_thread, self._keep_alive_thread]
            await asyncio.gather(*filter(None, tasks), return_exceptions=True)

        except asyncio.CancelledError as e:
            self.logger.error("tasks cancelled error: %s", e)

        if self._socket is not None:
            self.logger.notice("closing socket...")
            await self._socket.close()

        self._socket = None

        self.logger.notice("finish succeeded")
        self.logger.debug("AsyncLiveClient.finish LEAVE")
        return True


async def _socket_connect(websocket_url, headers) -> websockets.WebSocketClientProtocol:
    destination = websocket_url
    updated_headers = headers
    return await websockets.connect(
        destination, extra_headers=updated_headers, ping_interval=PING_INTERVAL
    )



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from urllib.parse import urlparse, urlunparse, parse_qs, urlencode
from typing import Dict


# appends query parameters to a URL
def append_query_params(url: str, params: Dict) -> str:
    parsed_url = urlparse(url)
    query_params = parse_qs(parsed_url.query)

    for key, value in params.items():
        if value is None:
            continue
        if isinstance(value, bool):
            value = str(value).lower()
        if isinstance(value, list):
            for item in value:
                query_params[key] = query_params.get(key, []) + [str(item)]
        else:
            query_params[key] = [str(value)]

    updated_query_string = urlencode(query_params, doseq=True)
    updated_url = parsed_url._replace(query=updated_query_string).geturl()
    return updated_url



# Copyright 2023-2024 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

import inspect
import asyncio
import threading
from typing import Optional
import logging, verboselogs

from .errors import DeepgramMicrophoneError
from .constants import LOGGING, CHANNELS, RATE, CHUNK


class Microphone:
    """
    This implements a microphone for local audio input. This uses PyAudio under the hood.
    """

    def __init__(
        self,
        push_callback,
        verbose: int = LOGGING,
        rate: int = RATE,
        chunk: int = CHUNK,
        channels: int = CHANNELS,
        input_device_index: Optional[int] = None,
    ):
        # dynamic import of pyaudio as not to force the requirements on the SDK (and users)
        import pyaudio

        self.logger = logging.getLogger(__name__)
        self.logger.addHandler(logging.StreamHandler())
        self.logger.setLevel(verbose)
        self.exit = threading.Event()

        self.audio = pyaudio.PyAudio()
        self.chunk = chunk
        self.rate = rate
        self.format = pyaudio.paInt16
        self.channels = channels
        self.input_device_index = input_device_index
        self.asyncio_loop = None
        self.asyncio_thread = None

        if inspect.iscoroutinefunction(push_callback):
            self.logger.verbose("async/await callback - wrapping")
            # Run our own asyncio loop.
            self.asyncio_thread = threading.Thread(target=self._start_asyncio_loop)
            self.asyncio_thread.start()

            self.push_callback = lambda data: asyncio.run_coroutine_threadsafe(
                push_callback(data), self.asyncio_loop
            ).result()
        else:
            self.logger.verbose("regular threaded callback")
            self.push_callback = push_callback

        self.stream = None

    def _start_asyncio_loop(self) -> None:
        self.asyncio_loop = asyncio.new_event_loop()
        self.asyncio_loop.run_forever()

    def is_active(self) -> bool:
        """
        returns True if the stream is active, False otherwise
        """
        self.logger.debug("Microphone.is_active ENTER")
        if self.stream is None:
            self.logger.error("stream is None")
            self.logger.debug("Microphone.is_active LEAVE")
            return False

        val = self.stream.is_active()
        self.logger.info("is_active: %s", val)
        self.logger.info("is_exiting: %s", self.exit.is_set())
        self.logger.debug("Microphone.is_active LEAVE")
        return val

    def start(self) -> bool:
        """
        starts the microphone stream
        """
        self.logger.debug("Microphone.start ENTER")

        if self.stream is not None:
            self.logger.error("start() failed. Library already initialized.")
            self.logger.debug("Microphone.start LEAVE")
            return False

        self.logger.info("format: %s", self.format)
        self.logger.info("channels: %d", self.channels)
        self.logger.info("rate: %d", self.rate)
        self.logger.info("chunk: %d", self.chunk)
        self.logger.info("input_device_id: %d", self.input_device_index)

        self.stream = self.audio.open(
            format=self.format,
            channels=self.channels,
            rate=self.rate,
            input=True,
            frames_per_buffer=self.chunk,
            input_device_index=self.input_device_index,
            stream_callback=self._callback,
        )

        self.exit.clear()
        self.stream.start_stream()

        self.logger.notice("start() succeeded")
        self.logger.debug("Microphone.start LEAVE")
        return True

    def _callback(self, input_data, frame_count, time_info, status_flags):
        """
        The callback used to process data in callback mode.
        """
        import pyaudio

        self.logger.debug("Microphone._callback ENTER")

        if self.exit.is_set():
            self.logger.info("exit is Set")
            self.logger.notice("_callback stopping...")
            self.logger.debug("Microphone._callback LEAVE")
            return None, pyaudio.paAbort

        if input_data is None:
            self.logger.warning("input_data is None")
            self.logger.debug("Microphone._callback LEAVE")
            return None, pyaudio.paContinue

        try:
            self.push_callback(input_data)
        except Exception as e:
            self.logger.error("Error while sending: %s", str(e))
            self.logger.debug("Microphone._callback LEAVE")
            raise

        self.logger.debug("Microphone._callback LEAVE")
        return input_data, pyaudio.paContinue

    def finish(self) -> bool:
        """
        Stops the microphone stream
        """
        self.logger.debug("Microphone.finish ENTER")

        self.logger.notice("signal exit")
        self.exit.set()

        if self.stream is not None:
            self.stream.stop_stream()
            self.stream.close()
        self.stream = None

        if self.asyncio_thread is not None:
            self.asyncio_loop.call_soon_threadsafe(self.asyncio_loop.stop)
            self.asyncio_thread.join()  # Clean up.
        self.asyncio_thread = None

        self.logger.notice("stream/recv thread joined")

        self.logger.notice("finish succeeded")
        self.logger.debug("Microphone.finish LEAVE")

        return True
